"""
Integration tests for the rabbitmq_plugin states
"""

import logging

import pytest

import salt.modules.rabbitmq as rabbitmq
import salt.states.rabbitmq_plugin as rabbitmq_plugin
from tests.support.mock import patch

log = logging.getLogger(__name__)

pytest.importorskip("docker")

pytestmark = [
    pytest.mark.slow_test,
    pytest.mark.skip_if_binaries_missing(
        "docker", "dockerd", reason="Docker not installed"
    ),
]


def mock_get_rabbitmq_plugin():
    return "/opt/rabbitmq/sbin/rabbitmq-plugins"


@pytest.fixture
def configure_loader_modules(docker_cmd_run_all_wrapper):
    return {
        rabbitmq_plugin: {
            "__salt__": {
                "rabbitmq.plugin_is_enabled": rabbitmq.plugin_is_enabled,
                "rabbitmq.enable_plugin": rabbitmq.enable_plugin,
                "rabbitmq.disable_plugin": rabbitmq.disable_plugin,
            },
            "__opts__": {"test": False},
            "_utils__": {},
        },
        rabbitmq: {
            "__salt__": {"cmd.run_all": docker_cmd_run_all_wrapper},
            "__opts__": {},
            "_utils__": {},
        },
    }


def test_enabled_enabled_disabled(rabbitmq_container):
    """
    Test rabbitmq_plugin.enabled and rabbitmq_plugin_disabled

    First enable the plugin.

    Second try to enable the plugin again.

    Third, try to disable the plugin.
    """

    with patch.object(rabbitmq, "_get_rabbitmq_plugin", mock_get_rabbitmq_plugin):
        # Enable the plugin
        ret = rabbitmq_plugin.enabled("rabbitmq_auth_backend_http")
        expected = {
            "name": "rabbitmq_auth_backend_http",
            "result": True,
            "comment": "Plugin 'rabbitmq_auth_backend_http' was enabled.",
            "changes": {"old": "", "new": "rabbitmq_auth_backend_http"},
        }
        assert ret == expected

        # Try to enable the plugin again
        ret = rabbitmq_plugin.enabled("rabbitmq_auth_backend_http")
        expected = {
            "name": "rabbitmq_auth_backend_http",
            "result": True,
            "comment": "Plugin 'rabbitmq_auth_backend_http' is already enabled.",
            "changes": {},
        }
        assert ret == expected

        # Disable the plugin
        ret = rabbitmq_plugin.disabled("rabbitmq_auth_backend_http")
        expected = {
            "name": "rabbitmq_auth_backend_http",
            "result": True,
            "comment": "Plugin 'rabbitmq_auth_backend_http' was disabled.",
            "changes": {"new": "", "old": "rabbitmq_auth_backend_http"},
        }
        assert ret == expected


def test_disabled(rabbitmq_container):
    """
    Test rabbitmq_plugin.enabled and rabbitmq_plugin_disabled

    First try to disable the plugin.

    Second enable the plugin again.

    Third disable the plugin.
    """

    with patch.object(rabbitmq, "_get_rabbitmq_plugin", mock_get_rabbitmq_plugin):
        # Try to disable the plugin
        ret = rabbitmq_plugin.disabled("rabbitmq_auth_backend_http")
        expected = {
            "name": "rabbitmq_auth_backend_http",
            "result": True,
            "comment": "Plugin 'rabbitmq_auth_backend_http' is already disabled.",
            "changes": {},
        }
        assert ret == expected

        # Enable the plugin
        ret = rabbitmq_plugin.enabled("rabbitmq_auth_backend_http")
        expected = {
            "name": "rabbitmq_auth_backend_http",
            "result": True,
            "comment": "Plugin 'rabbitmq_auth_backend_http' was enabled.",
            "changes": {"old": "", "new": "rabbitmq_auth_backend_http"},
        }
        assert ret == expected

        # Disable the plugin
        ret = rabbitmq_plugin.disabled("rabbitmq_auth_backend_http")
        expected = {
            "name": "rabbitmq_auth_backend_http",
            "result": True,
            "comment": "Plugin 'rabbitmq_auth_backend_http' was disabled.",
            "changes": {"new": "", "old": "rabbitmq_auth_backend_http"},
        }
        assert ret == expected



"""
Classes that manage file clients
"""

import contextlib
import errno
import ftplib  # nosec
import http.server
import logging
import os
import shutil
import string
import time
import urllib.error
import urllib.parse

import salt.channel.client
import salt.client
import salt.crypt
import salt.fileserver
import salt.loader
import salt.payload
import salt.utils.atomicfile
import salt.utils.data
import salt.utils.files
import salt.utils.gzip_util
import salt.utils.hashutils
import salt.utils.http
import salt.utils.path
import salt.utils.platform
import salt.utils.stringutils
import salt.utils.templates
import salt.utils.url
import salt.utils.verify
import salt.utils.versions
from salt.config import DEFAULT_HASH_TYPE
from salt.exceptions import CommandExecutionError, MinionError, SaltClientError
from salt.ext.tornado.httputil import (
    HTTPHeaders,
    HTTPInputError,
    parse_response_start_line,
)
from salt.utils.openstack.swift import SaltSwift

log = logging.getLogger(__name__)
MAX_FILENAME_LENGTH = 255


def get_file_client(opts, pillar=False):
    """
    Read in the ``file_client`` option and return the correct type of file
    server
    """
    client = opts.get("file_client", "remote")

    if pillar and client == "local":
        client = "pillar"
    return {"remote": RemoteClient, "local": FSClient, "pillar": PillarClient}.get(
        client, RemoteClient
    )(opts)


def decode_dict_keys_to_str(src):
    """
    Convert top level keys from bytes to strings if possible.
    This is necessary because Python 3 makes a distinction
    between these types.
    """
    if not isinstance(src, dict):
        return src

    output = {}
    for key, val in src.items():
        if isinstance(key, bytes):
            try:
                key = key.decode()
            except UnicodeError:
                pass
        output[key] = val
    return output


class Client:
    """
    Base class for Salt file interactions
    """

    def __init__(self, opts):
        self.opts = opts
        self.utils = salt.loader.utils(self.opts)

    # Add __setstate__ and __getstate__ so that the object may be
    # deep copied. It normally can't be deep copied because its
    # constructor requires an 'opts' parameter.
    # The TCP transport needs to be able to deep copy this class
    # due to 'salt.utils.context.ContextDict.clone'.
    def __setstate__(self, state):
        # This will polymorphically call __init__
        # in the derived class.
        self.__init__(state["opts"])

    def __getstate__(self):
        return {"opts": self.opts}

    def _check_proto(self, path):
        """
        Make sure that this path is intended for the salt master and trim it
        """
        if not path.startswith("salt://"):
            raise MinionError(f"Unsupported path: {path}")
        file_path, saltenv = salt.utils.url.parse(path)
        return file_path

    def _file_local_list(self, dest):
        """
        Helper util to return a list of files in a directory
        """
        if os.path.isdir(dest):
            destdir = dest
        else:
            destdir = os.path.dirname(dest)

        filelist = set()

        for root, dirs, files in salt.utils.path.os_walk(destdir, followlinks=True):
            for name in files:
                path = os.path.join(root, name)
                filelist.add(path)

        return filelist

    @contextlib.contextmanager
    def _cache_loc(self, path, saltenv="base", cachedir=None):
        """
        Return the local location to cache the file, cache dirs will be made
        """
        cachedir = self.get_cachedir(cachedir)
        dest = salt.utils.path.join(cachedir, "files", saltenv, path)
        destdir = os.path.dirname(dest)
        with salt.utils.files.set_umask(0o077):
            # remove destdir if it is a regular file to avoid an OSError when
            # running os.makedirs below
            if os.path.isfile(destdir):
                os.remove(destdir)

            # ensure destdir exists
            try:
                os.makedirs(destdir)
            except OSError as exc:
                if exc.errno != errno.EEXIST:  # ignore if it was there already
                    raise

            yield dest

    def get_cachedir(self, cachedir=None):
        if cachedir is None:
            cachedir = self.opts["cachedir"]
        elif not os.path.isabs(cachedir):
            cachedir = os.path.join(self.opts["cachedir"], cachedir)
        return cachedir

    def get_file(
        self, path, dest="", makedirs=False, saltenv="base", gzip=None, cachedir=None
    ):
        """
        Copies a file from the local files or master depending on
        implementation
        """
        raise NotImplementedError

    def file_list_emptydirs(self, saltenv="base", prefix=""):
        """
        List the empty dirs
        """
        raise NotImplementedError

    def cache_file(
        self,
        path,
        saltenv="base",
        cachedir=None,
        source_hash=None,
        verify_ssl=True,
        use_etag=False,
    ):
        """
        Pull a file down from the file server and store it in the minion
        file cache
        """
        return self.get_url(
            path,
            "",
            True,
            saltenv,
            cachedir=cachedir,
            source_hash=source_hash,
            verify_ssl=verify_ssl,
            use_etag=use_etag,
        )

    def cache_files(self, paths, saltenv="base", cachedir=None):
        """
        Download a list of files stored on the master and put them in the
        minion file cache
        """
        ret = []
        if isinstance(paths, str):
            paths = paths.split(",")
        for path in paths:
            ret.append(self.cache_file(path, saltenv, cachedir=cachedir))
        return ret

    def cache_master(self, saltenv="base", cachedir=None):
        """
        Download and cache all files on a master in a specified environment
        """
        ret = []
        for path in self.file_list(saltenv):
            ret.append(
                self.cache_file(salt.utils.url.create(path), saltenv, cachedir=cachedir)
            )
        return ret

    def cache_dir(
        self,
        path,
        saltenv="base",
        include_empty=False,
        include_pat=None,
        exclude_pat=None,
        cachedir=None,
    ):
        """
        Download all of the files in a subdir of the master
        """
        ret = []

        path = self._check_proto(salt.utils.data.decode(path))
        # We want to make sure files start with this *directory*, use
        # '/' explicitly because the master (that's generating the
        # list of files) only runs on POSIX
        if not path.endswith("/"):
            path = path + "/"

        log.info("Caching directory '%s' for environment '%s'", path, saltenv)
        # go through the list of all files finding ones that are in
        # the target directory and caching them
        for fn_ in self.file_list(saltenv):
            fn_ = salt.utils.data.decode(fn_)
            if fn_.strip() and fn_.startswith(path):
                if salt.utils.stringutils.check_include_exclude(
                    fn_, include_pat, exclude_pat
                ):
                    fn_ = self.cache_file(
                        salt.utils.url.create(fn_), saltenv, cachedir=cachedir
                    )
                    if fn_:
                        ret.append(fn_)

        if include_empty:
            # Break up the path into a list containing the bottom-level
            # directory (the one being recursively copied) and the directories
            # preceding it
            # separated = string.rsplit(path, '/', 1)
            # if len(separated) != 2:
            #     # No slashes in path. (So all files in saltenv will be copied)
            #     prefix = ''
            # else:
            #     prefix = separated[0]
            cachedir = self.get_cachedir(cachedir)

            dest = salt.utils.path.join(cachedir, "files", saltenv)
            for fn_ in self.file_list_emptydirs(saltenv):
                fn_ = salt.utils.data.decode(fn_)
                if fn_.startswith(path):
                    minion_dir = f"{dest}/{fn_}"
                    if not os.path.isdir(minion_dir):
                        os.makedirs(minion_dir)
                    ret.append(minion_dir)
        return ret

    def cache_local_file(self, path, **kwargs):
        """
        Cache a local file on the minion in the localfiles cache
        """
        dest = os.path.join(self.opts["cachedir"], "localfiles", path.lstrip("/"))
        destdir = os.path.dirname(dest)

        if not os.path.isdir(destdir):
            os.makedirs(destdir)

        shutil.copyfile(path, dest)
        return dest

    def file_local_list(self, saltenv="base"):
        """
        List files in the local minion files and localfiles caches
        """
        filesdest = os.path.join(self.opts["cachedir"], "files", saltenv)
        localfilesdest = os.path.join(self.opts["cachedir"], "localfiles")

        fdest = self._file_local_list(filesdest)
        ldest = self._file_local_list(localfilesdest)
        return sorted(fdest.union(ldest))

    def file_list(self, saltenv="base", prefix=""):
        """
        This function must be overwritten
        """
        return []  # pragma: no cover

    def dir_list(self, saltenv="base", prefix=""):
        """
        This function must be overwritten
        """
        return []  # pragma: no cover

    def symlink_list(self, saltenv="base", prefix=""):
        """
        This function must be overwritten
        """
        return {}  # pragma: no cover

    def is_cached(self, path, saltenv="base", cachedir=None):
        """
        Returns the full path to a file if it is cached locally on the minion
        otherwise returns a blank string
        """
        if path.startswith("salt://"):
            path, senv = salt.utils.url.parse(path)
            if senv:
                saltenv = senv

        escaped = True if salt.utils.url.is_escaped(path) else False

        # also strip escape character '|'
        localsfilesdest = os.path.join(
            self.opts["cachedir"], "localfiles", path.lstrip("|/")
        )
        filesdest = os.path.join(
            self.opts["cachedir"], "files", saltenv, path.lstrip("|/")
        )
        extrndest = self._extrn_path(path, saltenv, cachedir=cachedir)

        if os.path.exists(filesdest):
            return salt.utils.url.escape(filesdest) if escaped else filesdest
        elif os.path.exists(localsfilesdest):
            return (
                salt.utils.url.escape(localsfilesdest) if escaped else localsfilesdest
            )
        elif os.path.exists(extrndest):
            return extrndest

        return ""

    def cache_dest(self, url, saltenv="base", cachedir=None):
        """
        Return the expected cache location for the specified URL and
        environment.
        """
        proto = urllib.parse.urlparse(url).scheme

        if proto == "":
            # Local file path
            return url

        if proto == "salt":
            url, senv = salt.utils.url.parse(url)
            if senv:
                saltenv = senv
            return salt.utils.path.join(
                self.opts["cachedir"], "files", saltenv, url.lstrip("|/")
            )

        return self._extrn_path(url, saltenv, cachedir=cachedir)

    def list_states(self, saltenv):
        """
        Return a list of all available sls modules on the master for a given
        environment
        """
        states = set()
        for path in self.file_list(saltenv):
            if salt.utils.platform.is_windows():
                path = path.replace("\\", "/")
            if path.endswith(".sls"):
                # is an sls module!
                if path.endswith("/init.sls"):
                    states.add(path.replace("/", ".")[:-9])
                else:
                    states.add(path.replace("/", ".")[:-4])
        return sorted(states)

    def get_state(self, sls, saltenv, cachedir=None):
        """
        Get a state file from the master and store it in the local minion
        cache; return the location of the file
        """
        if "." in sls:
            sls = sls.replace(".", "/")
        sls_url = salt.utils.url.create(sls + ".sls")
        init_url = salt.utils.url.create(sls + "/init.sls")
        for path in [sls_url, init_url]:
            dest = self.cache_file(path, saltenv, cachedir=cachedir)
            if dest:
                return {"source": path, "dest": dest}
        return {}

    def get_dir(self, path, dest="", saltenv="base", gzip=None, cachedir=None):
        """
        Get a directory recursively from the salt-master
        """
        ret = []
        # Strip trailing slash
        path = self._check_proto(path).rstrip("/")
        # Break up the path into a list containing the bottom-level directory
        # (the one being recursively copied) and the directories preceding it
        separated = path.rsplit("/", 1)
        if len(separated) != 2:
            # No slashes in path. (This means all files in saltenv will be
            # copied)
            prefix = ""
        else:
            prefix = separated[0]

        # Copy files from master
        for fn_ in self.file_list(saltenv, prefix=path):
            # Prevent files in "salt://foobar/" (or salt://foo.sh) from
            # matching a path of "salt://foo"
            try:
                if fn_[len(path)] != "/":
                    continue
            except IndexError:
                continue
            # Remove the leading directories from path to derive
            # the relative path on the minion.
            minion_relpath = fn_[len(prefix) :].lstrip("/")
            ret.append(
                self.get_file(
                    salt.utils.url.create(fn_),
                    f"{dest}/{minion_relpath}",
                    True,
                    saltenv,
                    gzip,
                )
            )
        # Replicate empty dirs from master
        try:
            for fn_ in self.file_list_emptydirs(saltenv, prefix=path):
                # Prevent an empty dir "salt://foobar/" from matching a path of
                # "salt://foo"
                try:
                    if fn_[len(path)] != "/":
                        continue
                except IndexError:
                    continue
                # Remove the leading directories from path to derive
                # the relative path on the minion.
                minion_relpath = fn_[len(prefix) :].lstrip("/")
                minion_mkdir = f"{dest}/{minion_relpath}"
                if not os.path.isdir(minion_mkdir):
                    os.makedirs(minion_mkdir)
                ret.append(minion_mkdir)
        except TypeError:
            pass
        ret.sort()
        return ret

    def get_url(
        self,
        url,
        dest,
        makedirs=False,
        saltenv="base",
        no_cache=False,
        cachedir=None,
        source_hash=None,
        verify_ssl=True,
        use_etag=False,
    ):
        """
        Get a single file from a URL.
        """
        url_data = urllib.parse.urlparse(url, allow_fragments=False)
        url_scheme = url_data.scheme
        url_path = os.path.join(url_data.netloc, url_data.path).rstrip(os.sep)

        # If dest is a directory, rewrite dest with filename
        if dest is not None and (os.path.isdir(dest) or dest.endswith(("/", "\\"))):
            if (
                url_data.query
                or len(url_data.path) > 1
                and not url_data.path.endswith("/")
            ):
                strpath = url.split("/")[-1]
            else:
                strpath = "index.html"

            if salt.utils.platform.is_windows():
                strpath = salt.utils.path.sanitize_win_path(strpath)

            dest = os.path.join(dest, strpath)

        if url_scheme and url_scheme.lower() in string.ascii_lowercase:
            url_path = ":".join((url_scheme, url_path))
            url_scheme = "file"

        if url_scheme in ("file", ""):
            # Local filesystem
            if not os.path.isabs(url_path):
                raise CommandExecutionError(f"Path '{url_path}' is not absolute")
            if dest is None:
                with salt.utils.files.fopen(url_path, "rb") as fp_:
                    data = fp_.read()
                return data
            return url_path

        if url_scheme == "salt":
            result = self.get_file(url, dest, makedirs, saltenv, cachedir=cachedir)
            if result and dest is None:
                with salt.utils.files.fopen(result, "rb") as fp_:
                    data = fp_.read()
                return data
            return result

        if dest:
            destdir = os.path.dirname(dest)
            if not os.path.isdir(destdir):
                if makedirs:
                    os.makedirs(destdir)
                else:
                    return ""
        elif not no_cache:
            dest = self._extrn_path(url, saltenv, cachedir=cachedir)
            if source_hash is not None:
                try:
                    source_hash = source_hash.split("=")[-1]
                    form = salt.utils.files.HASHES_REVMAP[len(source_hash)]
                    if salt.utils.hashutils.get_hash(dest, form) == source_hash:
                        log.debug(
                            "Cached copy of %s (%s) matches source_hash %s, "
                            "skipping download",
                            url,
                            dest,
                            source_hash,
                        )
                        return dest
                except (AttributeError, KeyError, OSError):
                    pass
            destdir = os.path.dirname(dest)
            if not os.path.isdir(destdir):
                os.makedirs(destdir)

        if url_data.scheme == "s3":
            try:

                def s3_opt(key, default=None):
                    """
                    Get value of s3.<key> from Minion config or from Pillar
                    """
                    if "s3." + key in self.opts:
                        return self.opts["s3." + key]
                    try:
                        return self.opts["pillar"]["s3"][key]
                    except (KeyError, TypeError):
                        return default

                self.utils["s3.query"](
                    method="GET",
                    bucket=url_data.netloc,
                    path=url_data.path[1:],
                    return_bin=False,
                    local_file=dest,
                    action=None,
                    key=s3_opt("key"),
                    keyid=s3_opt("keyid"),
                    service_url=s3_opt("service_url"),
                    verify_ssl=s3_opt("verify_ssl", True),
                    location=s3_opt("location"),
                    path_style=s3_opt("path_style", False),
                    https_enable=s3_opt("https_enable", True),
                )
                return dest
            except Exception as exc:  # pylint: disable=broad-except
                raise MinionError(f"Could not fetch from {url}. Exception: {exc}")
        if url_data.scheme == "ftp":
            try:
                ftp = ftplib.FTP()  # nosec
                ftp_port = url_data.port
                if not ftp_port:
                    ftp_port = 21
                ftp.connect(url_data.hostname, ftp_port)
                ftp.login(url_data.username, url_data.password)
                remote_file_path = url_data.path.lstrip("/")
                with salt.utils.files.fopen(dest, "wb") as fp_:
                    ftp.retrbinary(f"RETR {remote_file_path}", fp_.write)
                ftp.quit()
                return dest
            except Exception as exc:  # pylint: disable=broad-except
                raise MinionError(
                    "Could not retrieve {} from FTP server. Exception: {}".format(
                        url, exc
                    )
                )

        if url_data.scheme == "swift":
            try:

                def swift_opt(key, default):
                    """
                    Get value of <key> from Minion config or from Pillar
                    """
                    if key in self.opts:
                        return self.opts[key]
                    try:
                        return self.opts["pillar"][key]
                    except (KeyError, TypeError):
                        return default

                swift_conn = SaltSwift(
                    swift_opt("keystone.user", None),
                    swift_opt("keystone.tenant", None),
                    swift_opt("keystone.auth_url", None),
                    swift_opt("keystone.password", None),
                )

                swift_conn.get_object(url_data.netloc, url_data.path[1:], dest)
                return dest
            except Exception:  # pylint: disable=broad-except
                raise MinionError(f"Could not fetch from {url}")

        get_kwargs = {}
        if url_data.username is not None and url_data.scheme in ("http", "https"):
            netloc = url_data.netloc
            at_sign_pos = netloc.rfind("@")
            if at_sign_pos != -1:
                netloc = netloc[at_sign_pos + 1 :]
            fixed_url = urllib.parse.urlunparse(
                (
                    url_data.scheme,
                    netloc,
                    url_data.path,
                    url_data.params,
                    url_data.query,
                    url_data.fragment,
                )
            )
            get_kwargs["auth"] = (url_data.username, url_data.password)
        else:
            fixed_url = url

        destfp = None
        dest_etag = f"{dest}.etag"
        try:
            # Tornado calls streaming_callback on redirect response bodies.
            # But we need streaming to support fetching large files (> RAM
            # avail). Here we are working around this by disabling recording
            # the body for redirections. The issue is fixed in Tornado 4.3.0
            # so on_header callback could be removed when we'll deprecate
            # Tornado<4.3.0. See #27093 and #30431 for details.

            # Use list here to make it writable inside the on_header callback.
            # Simple bool doesn't work here: on_header creates a new local
            # variable instead. This could be avoided in Py3 with 'nonlocal'
            # statement. There is no Py2 alternative for this.
            #
            # write_body[0] is used by the on_chunk callback to tell it whether
            #   or not we need to write the body of the request to disk. For
            #   30x redirects we set this to False because we don't want to
            #   write the contents to disk, as we will need to wait until we
            #   get to the redirected URL.
            #
            # write_body[1] will contain a tornado.httputil.HTTPHeaders
            #   instance that we will use to parse each header line. We
            #   initialize this to False, and after we parse the status line we
            #   will replace it with the HTTPHeaders instance. If/when we have
            #   found the encoding used in the request, we set this value to
            #   False to signify that we are done parsing.
            #
            # write_body[2] is where the encoding will be stored
            #
            # write_body[3] is where the etag will be stored if use_etag is
            #   enabled. This allows us to iterate over the headers until
            #   both content encoding and etag are found.
            write_body = [None, False, None, None]

            def on_header(hdr):
                if write_body[1] is not False and (
                    write_body[2] is None or (use_etag and write_body[3] is None)
                ):
                    if not hdr.strip() and "Content-Type" not in write_body[1]:
                        # If write_body[0] is True, then we are not following a
                        # redirect (initial response was a 200 OK). So there is
                        # no need to reset write_body[0].
                        if write_body[0] is not True:
                            # We are following a redirect, so we need to reset
                            # write_body[0] so that we properly follow it.
                            write_body[0] = None
                        # We don't need the HTTPHeaders object anymore
                        if not use_etag or write_body[3]:
                            write_body[1] = False
                        return
                    # Try to find out what content type encoding is used if
                    # this is a text file
                    write_body[1].parse_line(hdr)  # pylint: disable=no-member
                    # Case insensitive Etag header checking below. Don't break case
                    # insensitivity unless you really want to mess with people's heads
                    # in the tests. Note: http.server and apache2 use "Etag" and nginx
                    # uses "ETag" as the header key. Yay standards!
                    if use_etag and "etag" in map(str.lower, write_body[1]):
                        etag = write_body[3] = [
                            val
                            for key, val in write_body[1].items()
                            if key.lower() == "etag"
                        ][0]
                        with salt.utils.files.fopen(dest_etag, "w") as etagfp:
                            etag = etagfp.write(etag)
                    elif "Content-Type" in write_body[1]:
                        content_type = write_body[1].get(
                            "Content-Type"
                        )  # pylint: disable=no-member
                        if not content_type.startswith("text"):
                            write_body[2] = False
                            if not use_etag or write_body[3]:
                                write_body[1] = False
                        else:
                            encoding = "utf-8"
                            fields = content_type.split(";")
                            for field in fields:
                                if "encoding" in field:
                                    encoding = field.split("encoding=")[-1]
                            write_body[2] = encoding
                            # We have found our encoding. Stop processing headers.
                            if not use_etag or write_body[3]:
                                write_body[1] = False

                        # If write_body[0] is False, this means that this
                        # header is a 30x redirect, so we need to reset
                        # write_body[0] to None so that we parse the HTTP
                        # status code from the redirect target. Additionally,
                        # we need to reset write_body[2] so that we inspect the
                        # headers for the Content-Type of the URL we're
                        # following.
                        if write_body[0] is write_body[1] is False:
                            write_body[0] = write_body[2] = None

                # Check the status line of the HTTP request
                if write_body[0] is None:
                    try:
                        hdr = parse_response_start_line(hdr)
                    except HTTPInputError:
                        # Not the first line, do nothing
                        return
                    write_body[0] = hdr.code not in [301, 302, 303, 307]
                    write_body[1] = HTTPHeaders()

            if no_cache:
                result = []

                def on_chunk(chunk):
                    if write_body[0]:
                        if write_body[2]:
                            chunk = chunk.decode(write_body[2])
                        result.append(chunk)

            else:
                dest_tmp = f"{dest}.part"
                # We need an open filehandle to use in the on_chunk callback,
                # that's why we're not using a with clause here.
                # pylint: disable=resource-leakage
                destfp = salt.utils.files.fopen(dest_tmp, "wb")
                # pylint: enable=resource-leakage

                def on_chunk(chunk):
                    if write_body[0]:
                        destfp.write(chunk)

            # ETag is only used for refetch. Cached file and previous ETag
            # should be present for verification.
            header_dict = {}
            if use_etag and os.path.exists(dest_etag) and os.path.exists(dest):
                with salt.utils.files.fopen(dest_etag, "r") as etagfp:
                    etag = etagfp.read().replace("\n", "").strip()
                header_dict["If-None-Match"] = etag

            query = salt.utils.http.query(
                fixed_url,
                stream=True,
                streaming_callback=on_chunk,
                header_callback=on_header,
                username=url_data.username,
                password=url_data.password,
                opts=self.opts,
                verify_ssl=verify_ssl,
                header_dict=header_dict,
                **get_kwargs,
            )

            # 304 Not Modified is returned when If-None-Match header
            # matches server ETag for requested file.
            if use_etag and query.get("status") == 304:
                if not no_cache:
                    destfp.close()
                    destfp = None
                    os.remove(dest_tmp)
                return dest
            if "handle" not in query:
                raise MinionError(
                    "Error: {} reading {}".format(query["error"], url_data.path)
                )
            if no_cache:
                if write_body[2]:
                    return "".join(result)
                return b"".join(result)
            else:
                destfp.close()
                destfp = None
                salt.utils.files.rename(dest_tmp, dest)
                return dest
        except urllib.error.HTTPError as exc:
            raise MinionError(
                "HTTP error {0} reading {1}: {3}".format(
                    exc.code,
                    url,
                    *http.server.BaseHTTPRequestHandler.responses[exc.code],
                )
            )
        except urllib.error.URLError as exc:
            raise MinionError(f"Error reading {url}: {exc.reason}")
        finally:
            if destfp is not None:
                destfp.close()

    def get_template(
        self,
        url,
        dest,
        template="jinja",
        makedirs=False,
        saltenv="base",
        cachedir=None,
        **kwargs,
    ):
        """
        Cache a file then process it as a template
        """
        if "env" in kwargs:
            # "env" is not supported; Use "saltenv".
            kwargs.pop("env")

        kwargs["saltenv"] = saltenv
        sfn = self.cache_file(url, saltenv, cachedir=cachedir)
        if not sfn or not os.path.exists(sfn):
            return ""
        if template in salt.utils.templates.TEMPLATE_REGISTRY:
            data = salt.utils.templates.TEMPLATE_REGISTRY[template](sfn, **kwargs)
        else:
            log.error(
                "Attempted to render template with unavailable engine %s", template
            )
            return ""
        if not data["result"]:
            # Failed to render the template
            log.error("Failed to render template with error: %s", data["data"])
            return ""
        if not dest:
            # No destination passed, set the dest as an extrn_files cache
            dest = self._extrn_path(url, saltenv, cachedir=cachedir)
            # If Salt generated the dest name, create any required dirs
            makedirs = True

        destdir = os.path.dirname(dest)
        if not os.path.isdir(destdir):
            if makedirs:
                os.makedirs(destdir)
            else:
                salt.utils.files.safe_rm(data["data"])
                return ""
        shutil.move(data["data"], dest)
        return dest

    def _extrn_path(self, url, saltenv, cachedir=None):
        """
        Return the extrn_filepath for a given url
        """
        url_data = urllib.parse.urlparse(url)
        if salt.utils.platform.is_windows():
            netloc = salt.utils.path.sanitize_win_path(url_data.netloc)
        else:
            netloc = url_data.netloc

        # Strip user:pass from URLs
        netloc = netloc.split("@")[-1]
        try:
            if url_data.port:
                # Remove : from path
                netloc = netloc.replace(":", "")
        except ValueError:
            # On Windows urllib raises a ValueError
            # when using a file:// source and trying
            # to access the port attribute.
            pass

        if cachedir is None:
            cachedir = self.opts["cachedir"]
        elif not os.path.isabs(cachedir):
            cachedir = os.path.join(self.opts["cachedir"], cachedir)

        if url_data.query:
            file_name = "-".join([url_data.path, url_data.query])
        else:
            file_name = url_data.path

        # clean_path returns an empty string if the check fails
        root_path = salt.utils.path.join(cachedir, "extrn_files", saltenv, netloc)
        new_path = os.path.sep.join([root_path, file_name])
        if not salt.utils.verify.clean_path(root_path, new_path, subdir=True):
            return "Invalid path"

        if len(file_name) > MAX_FILENAME_LENGTH:
            file_name = salt.utils.hashutils.sha256_digest(file_name)

        return salt.utils.path.join(cachedir, "extrn_files", saltenv, netloc, file_name)


class PillarClient(Client):
    """
    Used by pillar to handle fileclient requests
    """

    def _find_file(self, path, saltenv="base"):
        """
        Locate the file path
        """
        fnd = {"path": "", "rel": ""}

        if salt.utils.url.is_escaped(path):
            # The path arguments are escaped
            path = salt.utils.url.unescape(path)
        for root in self.opts["pillar_roots"].get(saltenv, []):
            full = os.path.join(root, path)
            if os.path.isfile(full):
                fnd["path"] = full
                fnd["rel"] = path
                return fnd
        return fnd

    def get_file(
        self, path, dest="", makedirs=False, saltenv="base", gzip=None, cachedir=None
    ):
        """
        Copies a file from the local files directory into :param:`dest`
        gzip compression settings are ignored for local files
        """
        path = self._check_proto(path)
        fnd = self._find_file(path, saltenv)
        fnd_path = fnd.get("path")
        if not fnd_path:
            return ""

        return fnd_path

    def file_list(self, saltenv="base", prefix=""):
        """
        Return a list of files in the given environment
        with optional relative prefix path to limit directory traversal
        """
        ret = []
        prefix = prefix.strip("/")
        for path in self.opts["pillar_roots"].get(saltenv, []):
            for root, dirs, files in salt.utils.path.os_walk(
                os.path.join(path, prefix), followlinks=True
            ):
                # Don't walk any directories that match file_ignore_regex or glob
                dirs[:] = [
                    d for d in dirs if not salt.fileserver.is_file_ignored(self.opts, d)
                ]
                for fname in files:
                    relpath = os.path.relpath(os.path.join(root, fname), path)
                    ret.append(salt.utils.data.decode(relpath))
        return ret

    def file_list_emptydirs(self, saltenv="base", prefix=""):
        """
        List the empty dirs in the pillar_roots
        with optional relative prefix path to limit directory traversal
        """
        ret = []
        prefix = prefix.strip("/")
        for path in self.opts["pillar_roots"].get(saltenv, []):
            for root, dirs, files in salt.utils.path.os_walk(
                os.path.join(path, prefix), followlinks=True
            ):
                # Don't walk any directories that match file_ignore_regex or glob
                dirs[:] = [
                    d for d in dirs if not salt.fileserver.is_file_ignored(self.opts, d)
                ]
                if not dirs and not files:
                    ret.append(salt.utils.data.decode(os.path.relpath(root, path)))
        return ret

    def dir_list(self, saltenv="base", prefix=""):
        """
        List the dirs in the pillar_roots
        with optional relative prefix path to limit directory traversal
        """
        ret = []
        prefix = prefix.strip("/")
        for path in self.opts["pillar_roots"].get(saltenv, []):
            for root, dirs, files in salt.utils.path.os_walk(
                os.path.join(path, prefix), followlinks=True
            ):
                ret.append(salt.utils.data.decode(os.path.relpath(root, path)))
        return ret

    def __get_file_path(self, path, saltenv="base"):
        """
        Return either a file path or the result of a remote find_file call.
        """
        try:
            path = self._check_proto(path)
        except MinionError as err:
            # Local file path
            if not os.path.isfile(path):
                log.warning(
                    "specified file %s is not present to generate hash: %s", path, err
                )
                return None
            else:
                return path
        return self._find_file(path, saltenv)

    def hash_file(self, path, saltenv="base"):
        """
        Return the hash of a file, to get the hash of a file in the pillar_roots
        prepend the path with salt://<file on server> otherwise, prepend the
        file with / for a local file.
        """
        ret = {}
        fnd = self.__get_file_path(path, saltenv)
        if fnd is None:
            return ret

        try:
            # Remote file path (self._find_file() invoked)
            fnd_path = fnd["path"]
        except TypeError:
            # Local file path
            fnd_path = fnd

        hash_type = self.opts.get("hash_type", DEFAULT_HASH_TYPE)
        ret["hsum"] = salt.utils.hashutils.get_hash(fnd_path, form=hash_type)
        ret["hash_type"] = hash_type
        return ret

    def hash_and_stat_file(self, path, saltenv="base"):
        """
        Return the hash of a file, to get the hash of a file in the pillar_roots
        prepend the path with salt://<file on server> otherwise, prepend the
        file with / for a local file.

        Additionally, return the stat result of the file, or None if no stat
        results were found.
        """
        ret = {}
        fnd = self.__get_file_path(path, saltenv)
        if fnd is None:
            return ret, None

        try:
            # Remote file path (self._find_file() invoked)
            fnd_path = fnd["path"]
            fnd_stat = fnd.get("stat")
        except TypeError:
            # Local file path
            fnd_path = fnd
            try:
                fnd_stat = list(os.stat(fnd_path))
            except Exception:  # pylint: disable=broad-except
                fnd_stat = None

        hash_type = self.opts.get("hash_type", DEFAULT_HASH_TYPE)
        ret["hsum"] = salt.utils.hashutils.get_hash(fnd_path, form=hash_type)
        ret["hash_type"] = hash_type
        return ret, fnd_stat

    def list_env(self, saltenv="base"):
        """
        Return a list of the files in the file server's specified environment
        """
        return self.file_list(saltenv)

    def master_opts(self):
        """
        Return the master opts data
        """
        return self.opts

    def envs(self):
        """
        Return the available environments
        """
        ret = []
        for saltenv in self.opts["pillar_roots"]:
            ret.append(saltenv)
        return ret

    def master_tops(self):
        """
        Originally returned information via the external_nodes subsystem.
        External_nodes was deprecated and removed in
        2014.1.6 in favor of master_tops (which had been around since pre-0.17).
             salt-call --local state.show_top
        ends up here, but master_tops has not been extended to support
        show_top in a completely local environment yet.  It's worth noting
        that originally this fn started with
            if 'external_nodes' not in opts: return {}
        So since external_nodes is gone now, we are just returning the
        empty dict.
        """
        return {}


class RemoteClient(Client):
    """
    Interact with the salt master file server.
    """

    def __init__(self, opts):
        Client.__init__(self, opts)
        self._closing = False
        self.channel = salt.channel.client.ReqChannel.factory(self.opts)
        if hasattr(self.channel, "auth"):
            self.auth = self.channel.auth
        else:
            self.auth = ""

    def _refresh_channel(self):
        """
        Reset the channel, in the event of an interruption
        """
        # Close the previous channel
        self.channel.close()
        # Instantiate a new one
        self.channel = salt.channel.client.ReqChannel.factory(self.opts)
        return self.channel

    def _channel_send(self, load, raw=False):
        start = time.monotonic()
        try:
            return self.channel.send(
                load,
                raw=raw,
            )
        except salt.exceptions.SaltReqTimeoutError:
            raise SaltClientError(
                f"File client timed out after {int(time.monotonic() - start)} seconds"
            )

    def destroy(self):
        if self._closing:
            return

        self._closing = True
        channel = None
        try:
            channel = self.channel
        except AttributeError:
            pass
        if channel is not None:
            channel.close()

    def get_file(
        self, path, dest="", makedirs=False, saltenv="base", gzip=None, cachedir=None
    ):
        """
        Get a single file from the salt-master
        path must be a salt server location, aka, salt://path/to/file, if
        dest is omitted, then the downloaded file will be placed in the minion
        cache
        """
        path, senv = salt.utils.url.split_env(path)
        if senv:
            saltenv = senv

        if not salt.utils.platform.is_windows():
            hash_server, stat_server = self.hash_and_stat_file(path, saltenv)
        else:
            hash_server = self.hash_file(path, saltenv)

        # Check if file exists on server, before creating files and
        # directories
        if hash_server == "":
            log.debug("Could not find file '%s' in saltenv '%s'", path, saltenv)
            return False

        # If dest is a directory, rewrite dest with filename
        if dest is not None and (os.path.isdir(dest) or dest.endswith(("/", "\\"))):
            dest = os.path.join(dest, os.path.basename(path))
            log.debug(
                "In saltenv '%s', '%s' is a directory. Changing dest to '%s'",
                saltenv,
                os.path.dirname(dest),
                dest,
            )

        # Hash compare local copy with master and skip download
        # if no difference found.
        dest2check = dest
        if not dest2check:
            rel_path = self._check_proto(path)

            log.debug(
                "In saltenv '%s', looking at rel_path '%s' to resolve '%s'",
                saltenv,
                rel_path,
                path,
            )
            with self._cache_loc(rel_path, saltenv, cachedir=cachedir) as cache_dest:
                dest2check = cache_dest

        log.debug(
            "In saltenv '%s', ** considering ** path '%s' to resolve '%s'",
            saltenv,
            dest2check,
            path,
        )

        if dest2check and os.path.isfile(dest2check):
            if not salt.utils.platform.is_windows():
                hash_local, stat_local = self.hash_and_stat_file(dest2check, saltenv)
            else:
                hash_local = self.hash_file(dest2check, saltenv)

            if hash_local == hash_server:
                return dest2check

        log.debug(
            "Fetching file from saltenv '%s', ** attempting ** '%s'", saltenv, path
        )
        d_tries = 0
        transport_tries = 0
        path = self._check_proto(path)
        load = {"path": path, "saltenv": saltenv, "cmd": "_serve_file"}
        if gzip:
            gzip = int(gzip)
            load["gzip"] = gzip

        fn_ = None
        if dest:
            destdir = os.path.dirname(dest)
            if not os.path.isdir(destdir):
                if makedirs:
                    try:
                        os.makedirs(destdir)
                    except OSError as exc:
                        if exc.errno != errno.EEXIST:  # ignore if it was there already
                            raise
                else:
                    return False
            # We need an open filehandle here, that's why we're not using a
            # with clause:
            # pylint: disable=resource-leakage
            fn_ = salt.utils.files.fopen(dest, "wb+")
            # pylint: enable=resource-leakage
        else:
            log.debug("No dest file found")

        while True:
            if not fn_:
                load["loc"] = 0
            else:
                load["loc"] = fn_.tell()
            data = self._channel_send(
                load,
                raw=True,
            )
            # Sometimes the source is local (eg when using
            # 'salt.fileserver.FSChan'), in which case the keys are
            # already strings. Sometimes the source is remote, in which
            # case the keys are bytes due to raw mode. Standardize on
            # strings for the top-level keys to simplify things.
            data = decode_dict_keys_to_str(data)
            try:
                if not data["data"]:
                    if not fn_ and data["dest"]:
                        # This is a 0 byte file on the master
                        with self._cache_loc(
                            data["dest"], saltenv, cachedir=cachedir
                        ) as cache_dest:
                            dest = cache_dest
                            with salt.utils.files.fopen(cache_dest, "wb+") as ofile:
                                ofile.write(data["data"])
                    if "hsum" in data and d_tries < 3:
                        # Master has prompted a file verification, if the
                        # verification fails, re-download the file. Try 3 times
                        d_tries += 1
                        hsum = salt.utils.hashutils.get_hash(
                            dest,
                            salt.utils.stringutils.to_str(
                                data.get("hash_type", DEFAULT_HASH_TYPE)
                            ),
                        )
                        if hsum != data["hsum"]:
                            log.warning(
                                "Bad download of file %s, attempt %d of 3",
                                path,
                                d_tries,
                            )
                            continue
                    break
                if not fn_:
                    with self._cache_loc(
                        data["dest"], saltenv, cachedir=cachedir
                    ) as cache_dest:
                        dest = cache_dest
                        # If a directory was formerly cached at this path, then
                        # remove it to avoid a traceback trying to write the file
                        if os.path.isdir(dest):
                            salt.utils.files.rm_rf(dest)
                        fn_ = salt.utils.atomicfile.atomic_open(dest, "wb+")
                if data.get("gzip", None):
                    data = salt.utils.gzip_util.uncompress(data["data"])
                else:
                    data = data["data"]
                if isinstance(data, str):
                    data = data.encode()
                fn_.write(data)
            except (TypeError, KeyError) as exc:
                try:
                    data_type = type(data).__name__
                except AttributeError:
                    # Shouldn't happen, but don't let this cause a traceback.
                    data_type = str(type(data))
                transport_tries += 1
                log.warning(
                    "Data transport is broken, got: %s, type: %s, "
                    "exception: %s, attempt %d of 3",
                    data,
                    data_type,
                    exc,
                    transport_tries,
                )
                self._refresh_channel()
                if transport_tries > 3:
                    log.error(
                        "Data transport is broken, got: %s, type: %s, "
                        "exception: %s, retry attempts exhausted",
                        data,
                        data_type,
                        exc,
                    )
                    break

        if fn_:
            fn_.close()
            log.info("Fetching file from saltenv '%s', ** done ** '%s'", saltenv, path)
        else:
            log.debug(
                "In saltenv '%s', we are ** missing ** the file '%s'", saltenv, path
            )

        return dest

    def file_list(self, saltenv="base", prefix=""):
        """
        List the files on the master
        """
        load = {"saltenv": saltenv, "prefix": prefix, "cmd": "_file_list"}
        return self._channel_send(
            load,
        )

    def file_list_emptydirs(self, saltenv="base", prefix=""):
        """
        List the empty dirs on the master
        """
        load = {"saltenv": saltenv, "prefix": prefix, "cmd": "_file_list_emptydirs"}
        return self._channel_send(
            load,
        )

    def dir_list(self, saltenv="base", prefix=""):
        """
        List the dirs on the master
        """
        load = {"saltenv": saltenv, "prefix": prefix, "cmd": "_dir_list"}
        return self._channel_send(
            load,
        )

    def symlink_list(self, saltenv="base", prefix=""):
        """
        List symlinked files and dirs on the master
        """
        load = {"saltenv": saltenv, "prefix": prefix, "cmd": "_symlink_list"}
        return self._channel_send(
            load,
        )

    def __hash_and_stat_file(self, path, saltenv="base"):
        """
        Common code for hashing and stating files
        """
        try:
            path = self._check_proto(path)
        except MinionError as err:
            if not os.path.isfile(path):
                log.warning(
                    "specified file %s is not present to generate hash: %s", path, err
                )
                return {}, None
            else:
                ret = {}
                hash_type = self.opts.get("hash_type", DEFAULT_HASH_TYPE)
                ret["hsum"] = salt.utils.hashutils.get_hash(path, form=hash_type)
                ret["hash_type"] = hash_type
                return ret
        load = {"path": path, "saltenv": saltenv, "cmd": "_file_hash"}
        return self._channel_send(
            load,
        )

    def hash_file(self, path, saltenv="base"):
        """
        Return the hash of a file, to get the hash of a file on the salt
        master file server prepend the path with salt://<file on server>
        otherwise, prepend the file with / for a local file.
        """
        return self.__hash_and_stat_file(path, saltenv)

    def hash_and_stat_file(self, path, saltenv="base"):
        """
        The same as hash_file, but also return the file's mode, or None if no
        mode data is present.
        """
        hash_result = self.hash_file(path, saltenv)
        try:
            path = self._check_proto(path)
        except MinionError as err:
            if not os.path.isfile(path):
                return hash_result, None
            else:
                try:
                    return hash_result, list(os.stat(path))
                except Exception:  # pylint: disable=broad-except
                    return hash_result, None
        load = {"path": path, "saltenv": saltenv, "cmd": "_file_find"}
        fnd = self._channel_send(
            load,
        )
        try:
            stat_result = fnd.get("stat")
        except AttributeError:
            stat_result = None
        return hash_result, stat_result

    def list_env(self, saltenv="base"):
        """
        Return a list of the files in the file server's specified environment
        """
        load = {"saltenv": saltenv, "cmd": "_file_list"}
        return self._channel_send(
            load,
        )

    def envs(self):
        """
        Return a list of available environments
        """
        load = {"cmd": "_file_envs"}
        return self._channel_send(
            load,
        )

    def master_opts(self):
        """
        Return the master opts data
        """
        load = {"cmd": "_master_opts"}
        return self._channel_send(
            load,
        )

    def master_tops(self):
        """
        Return the metadata derived from the master_tops system
        """
        load = {"cmd": "_master_tops", "id": self.opts["id"], "opts": self.opts}
        return self._channel_send(
            load,
        )

    def __enter__(self):
        return self

    def __exit__(self, *args):
        self.destroy()


class FSClient(RemoteClient):
    """
    A local client that uses the RemoteClient but substitutes the channel for
    the FSChan object
    """

    def __init__(self, opts):  # pylint: disable=W0231
        Client.__init__(self, opts)  # pylint: disable=W0233
        self._closing = False
        self.channel = salt.fileserver.FSChan(opts)
        self.auth = DumbAuth()


# Provide backward compatibility for anyone directly using LocalClient (but no
# one should be doing this).
LocalClient = FSClient


class DumbAuth:
    """
    The dumbauth class is used to stub out auth calls fired from the FSClient
    subsystem
    """

    def gen_token(self, clear_tok):
        return clear_tok


class ContextlessFileClient:
    def __init__(self, file_client):
        self.file_client = file_client

    def __getattr__(self, key):
        return getattr(self.file_client, key)

    def __exit__(self, *_):
        pass

    def __enter__(self):
        return self



from collections import OrderedDict

import pytest

import salt.modules.restconf as restconf
from tests.support.mock import MagicMock, patch


@pytest.fixture
def configure_loader_modules():
    return {restconf: {}}


@pytest.fixture
def mocking_dunder_salt_restconf_getdata_response_always_200():
    fake_response = OrderedDict()
    fake_response["fjord"] = "meow"
    fake_restconf_response = dict(status=200, dict=fake_response)
    with patch.dict(
        restconf.__salt__,
        {"restconf.get_data": MagicMock(return_value=fake_restconf_response)},
    ) as opt_dunder_mock:
        yield opt_dunder_mock


def test_module_path_check_primary_success(
    mocking_dunder_salt_restconf_getdata_response_always_200,
):
    result = restconf.path_check("fakeprimarypath", "fakeinitpath")

    assert type(result) is dict
    assert "request_restponse" in result
    assert "fjord" in result["request_restponse"]
    assert "request_path" in result
    assert "request_path" in result
    assert "path_used" in result
    assert "result" in result
    assert result["result"] is True
    assert result["request_path"] == "fakeprimarypath"
    assert result["path_used"] == "primary"


@pytest.fixture
def mocking_dunder_salt_restconf_getdata_response_always_404():
    fake_restconf_response = dict(status=404)
    with patch.dict(
        restconf.__salt__,
        {"restconf.get_data": MagicMock(return_value=fake_restconf_response)},
    ) as opt_dunder_mock:
        yield opt_dunder_mock


def test_module_path_check_always_fail(
    mocking_dunder_salt_restconf_getdata_response_always_404,
):
    result = restconf.path_check("fakeprimarypath", "fakeinitpath")

    assert result["result"] is False
    assert type(result) is dict


@pytest.fixture
def mocking_dunder_salt_restconf_getdata_response_first404_then_200():
    fake_response = OrderedDict()
    fake_response["fjord"] = "meow"
    fake_restconf_response_1 = dict(status=404)
    fake_restconf_response_2 = dict(status=200, dict=fake_response)
    with patch.dict(
        restconf.__salt__,
        {
            "restconf.get_data": MagicMock(
                side_effect=[fake_restconf_response_1, fake_restconf_response_2]
            )
        },
    ) as opt_dunder_mock:
        yield opt_dunder_mock


def test_module_path_check_primary_init(
    mocking_dunder_salt_restconf_getdata_response_first404_then_200,
):
    result = restconf.path_check("fakeprimarypath", "fakeinitpath")

    assert type(result) is dict
    assert "request_restponse" in result
    assert "fjord" in result["request_restponse"]
    assert "request_path" in result
    assert "request_path" in result
    assert "path_used" in result
    assert "result" in result
    assert result["result"] is True
    assert result["request_path"] == "fakeinitpath"
    assert result["path_used"] == "init"



"""
    :codeauthor: Nicole Thomas <nicole@saltstack.com>
"""

import pytest

import salt.modules.mac_user as mac_user
from salt.exceptions import CommandExecutionError, SaltInvocationError
from tests.support.mock import MagicMock, patch

pwd = pytest.importorskip("pwd")


@pytest.fixture
def configure_loader_modules():
    return {mac_user: {}}


@pytest.fixture
def mock_pwall():
    return [
        pwd.struct_passwd(
            (
                "_amavisd",
                "*",
                83,
                83,
                "AMaViS Daemon",
                "/var/virusmails",
                "/usr/bin/false",
            )
        ),
        pwd.struct_passwd(
            (
                "_appleevents",
                "*",
                55,
                55,
                "AppleEvents Daemon",
                "/var/empty",
                "/usr/bin/false",
            )
        ),
        pwd.struct_passwd(
            (
                "_appowner",
                "*",
                87,
                87,
                "Application Owner",
                "/var/empty",
                "/usr/bin/false",
            )
        ),
    ]


@pytest.fixture
def mock_info_ret():
    return {
        "shell": "/bin/bash",
        "name": "test",
        "gid": 4376,
        "groups": ["TEST_GROUP"],
        "home": "/Users/foo",
        "fullname": "TEST USER",
        "uid": 4376,
    }


@pytest.mark.skip(reason="Waiting on some clarifications from bug report #10594")
def _test_flush_dscl_cache():
    # TODO: Implement tests after clarifications come in
    pass


def test_dscl():
    """
    Tests the creation of a dscl node
    """
    mac_mock = MagicMock(
        return_value={"pid": 4948, "retcode": 0, "stderr": "", "stdout": ""}
    )
    with patch.dict(mac_user.__salt__, {"cmd.run_all": mac_mock}):
        with patch.dict(
            mac_user.__grains__,
            {"kernel": "Darwin", "osrelease": "10.9.1", "osrelease_info": (10, 9, 1)},
        ):
            assert mac_user._dscl(["username", "UniqueID", 501]) == {
                "pid": 4948,
                "retcode": 0,
                "stderr": "",
                "stdout": "",
            }


def test_first_avail_uid(mock_pwall):
    """
    Tests the availability of the next uid
    """
    with patch("pwd.getpwall", MagicMock(return_value=mock_pwall)):
        assert mac_user._first_avail_uid() == 501


# 'add' function tests: 4
# Only tested error handling
# Full functionality tests covered in integration testing


def test_add_user_exists(mock_info_ret):
    """
    Tests if the user exists or not
    """
    with patch("salt.modules.mac_user.info", MagicMock(return_value=mock_info_ret)):
        pytest.raises(CommandExecutionError, mac_user.add, "test")


def test_add_whitespace():
    """
    Tests if there is whitespace in the user name
    """
    with patch("salt.modules.mac_user.info", MagicMock(return_value={})):
        pytest.raises(SaltInvocationError, mac_user.add, "foo bar")


def test_add_uid_int():
    """
    Tests if the uid is an int
    """
    with patch("salt.modules.mac_user.info", MagicMock(return_value={})):
        pytest.raises(SaltInvocationError, mac_user.add, "foo", "foo")


def test_add_gid_int():
    """
    Tests if the gid is an int
    """
    with patch("salt.modules.mac_user.info", MagicMock(return_value={})):
        pytest.raises(SaltInvocationError, mac_user.add, "foo", 20, "foo")


# 'delete' function tests: 2
# Only tested pure logic of function
# Full functionality tests covered in integration testing


def test_delete_whitespace():
    """
    Tests if there is whitespace in the user name
    """
    pytest.raises(SaltInvocationError, mac_user.delete, "foo bar")


def test_delete_user_exists():
    """
    Tests if the user exists or not
    """
    with patch("salt.modules.mac_user.info", MagicMock(return_value={})):
        assert mac_user.delete("foo")


def test_getent(mock_pwall):
    """
    Tests the list of information for all users
    """
    with patch("pwd.getpwall", MagicMock(return_value=mock_pwall)), patch(
        "salt.modules.mac_user.list_groups", MagicMock(return_value=["TEST_GROUP"])
    ):
        ret = [
            {
                "shell": "/usr/bin/false",
                "name": "_amavisd",
                "gid": 83,
                "groups": ["TEST_GROUP"],
                "home": "/var/virusmails",
                "fullname": "AMaViS Daemon",
                "uid": 83,
            },
            {
                "shell": "/usr/bin/false",
                "name": "_appleevents",
                "gid": 55,
                "groups": ["TEST_GROUP"],
                "home": "/var/empty",
                "fullname": "AppleEvents Daemon",
                "uid": 55,
            },
            {
                "shell": "/usr/bin/false",
                "name": "_appowner",
                "gid": 87,
                "groups": ["TEST_GROUP"],
                "home": "/var/empty",
                "fullname": "Application Owner",
                "uid": 87,
            },
        ]
        assert mac_user.getent() == ret


# 'chuid' function tests: 3
# Only tested pure logic of function
# Full functionality tests covered in integration testing


def test_chuid_int():
    """
    Tests if the uid is an int
    """
    pytest.raises(SaltInvocationError, mac_user.chuid, "foo", "foo")


def test_chuid_user_exists():
    """
    Tests if the user exists or not
    """
    with patch("salt.modules.mac_user.info", MagicMock(return_value={})):
        pytest.raises(CommandExecutionError, mac_user.chuid, "foo", 4376)


def test_chuid_same_uid(mock_info_ret):
    """
    Tests if the user's uid is the same as as the argument
    """
    with patch("salt.modules.mac_user.info", MagicMock(return_value=mock_info_ret)):
        assert mac_user.chuid("foo", 4376)


# 'chgid' function tests: 3
# Only tested pure logic of function
# Full functionality tests covered in integration testing


def test_chgid_int():
    """
    Tests if the gid is an int
    """
    pytest.raises(SaltInvocationError, mac_user.chgid, "foo", "foo")


def test_chgid_user_exists():
    """
    Tests if the user exists or not
    """
    with patch("salt.modules.mac_user.info", MagicMock(return_value={})):
        pytest.raises(CommandExecutionError, mac_user.chgid, "foo", 4376)


def test_chgid_same_gid(mock_info_ret):
    """
    Tests if the user's gid is the same as as the argument
    """
    with patch("salt.modules.mac_user.info", MagicMock(return_value=mock_info_ret)):
        assert mac_user.chgid("foo", 4376)


# 'chshell' function tests: 2
# Only tested pure logic of function
# Full functionality tests covered in integration testing


def test_chshell_user_exists():
    """
    Tests if the user exists or not
    """
    with patch("salt.modules.mac_user.info", MagicMock(return_value={})):
        pytest.raises(CommandExecutionError, mac_user.chshell, "foo", "/bin/bash")


def test_chshell_same_shell(mock_info_ret):
    """
    Tests if the user's shell is the same as the argument
    """
    with patch("salt.modules.mac_user.info", MagicMock(return_value=mock_info_ret)):
        assert mac_user.chshell("foo", "/bin/bash")


# 'chhome' function tests: 2
# Only tested pure logic of function
# Full functionality tests covered in integration testing


def test_chhome_user_exists():
    """
    Test if the user exists or not
    """
    with patch("salt.modules.mac_user.info", MagicMock(return_value={})):
        pytest.raises(CommandExecutionError, mac_user.chhome, "foo", "/Users/foo")


def test_chhome_same_home(mock_info_ret):
    """
    Tests if the user's home is the same as the argument
    """
    with patch("salt.modules.mac_user.info", MagicMock(return_value=mock_info_ret)):
        assert mac_user.chhome("foo", "/Users/foo")


# 'chfullname' function tests: 2
# Only tested pure logic of function
# Full functionality tests covered in integration testing


def test_chfullname_user_exists():
    """
    Tests if the user exists or not
    """
    with patch("salt.modules.mac_user.info", MagicMock(return_value={})):
        pytest.raises(CommandExecutionError, mac_user.chfullname, "test", "TEST USER")


def test_chfullname_same_name(mock_info_ret):
    """
    Tests if the user's full name is the same as the argument
    """
    with patch("salt.modules.mac_user.info", MagicMock(return_value=mock_info_ret)):
        assert mac_user.chfullname("test", "TEST USER")


# 'chgroups' function tests: 3
# Only tested pure logic of function
# Full functionality tests covered in integration testing


def test_chgroups_user_exists():
    """
    Tests if the user exists or not
    """
    with patch("salt.modules.mac_user.info", MagicMock(return_value={})):
        pytest.raises(CommandExecutionError, mac_user.chgroups, "foo", "wheel,root")


def test_chgroups_bad_groups(mock_info_ret):
    """
    Test if there is white space in groups argument
    """
    with patch("salt.modules.mac_user.info", MagicMock(return_value=mock_info_ret)):
        pytest.raises(SaltInvocationError, mac_user.chgroups, "test", "bad group")


def test_chgroups_same_desired(mock_info_ret):
    """
    Tests if the user's list of groups is the same as the arguments
    """
    mock_primary = MagicMock(return_value="wheel")
    with patch.dict(mac_user.__salt__, {"file.gid_to_group": mock_primary}), patch(
        "salt.modules.mac_user.info", MagicMock(return_value=mock_info_ret)
    ), patch(
        "salt.modules.mac_user.list_groups", MagicMock(return_value=("wheel", "root"))
    ):
        assert mac_user.chgroups("test", "wheel,root")


def test_info():
    """
    Tests the return of user information
    """
    mock_pwnam = pwd.struct_passwd(
        ("root", "*", 0, 0, "TEST USER", "/var/test", "/bin/bash")
    )
    ret = {
        "shell": "/bin/bash",
        "name": "root",
        "gid": 0,
        "groups": ["_TEST_GROUP"],
        "home": "/var/test",
        "fullname": "TEST USER",
        "uid": 0,
    }
    with patch("pwd.getpwall", MagicMock(return_value=[mock_pwnam])), patch(
        "salt.modules.mac_user.list_groups", MagicMock(return_value=["_TEST_GROUP"])
    ):
        assert mac_user.info("root") == ret


def test_format_info():
    """
    Tests the formatting of returned user information
    """
    data = pwd.struct_passwd(
        (
            "_TEST_GROUP",
            "*",
            83,
            83,
            "AMaViS Daemon",
            "/var/virusmails",
            "/usr/bin/false",
        )
    )
    ret = {
        "shell": "/usr/bin/false",
        "name": "_TEST_GROUP",
        "gid": 83,
        "groups": ["_TEST_GROUP"],
        "home": "/var/virusmails",
        "fullname": "AMaViS Daemon",
        "uid": 83,
    }
    with patch(
        "salt.modules.mac_user.list_groups", MagicMock(return_value=["_TEST_GROUP"])
    ):
        assert mac_user._format_info(data) == ret


def test_list_users():
    """
    Tests the list of all users
    """
    expected = ["spongebob", "patrick", "squidward"]
    mock_run = MagicMock(
        return_value={
            "pid": 4948,
            "retcode": 0,
            "stderr": "",
            "stdout": "\n".join(expected),
        }
    )
    with patch.dict(mac_user.__grains__, {"osrelease_info": (10, 9, 1)}), patch.dict(
        mac_user.__salt__, {"cmd.run_all": mock_run}
    ):
        assert mac_user.list_users() == expected


def test_kcpassword():
    hashes = {
        # Actual hashes from macOS, since reference implementation didn't account for trailing null
        "0": "4d 89 f9 91 1f 7a 46 5e f7 a8 11 ff",
        "password": "0d e8 21 50 a5 d3 af 8e a3 de d9 14",
        "shorterpwd": "0e e1 3d 51 a6 d9 af 9a d4 dd 1f 27",
        "Squarepants": "2e f8 27 42 a0 d9 ad 8b cd cd 6c 7d",
        "longerpasswd": "11 e6 3c 44 b7 ce ad 8b d0 ca 68 19 89 b1 65 ae 7e 89 12 b8 51 f8 f0 ff",
        "ridiculouslyextendedpass": "0f e0 36 4a b1 c9 b1 85 d6 ca 73 04 ec 2a 57 b7 d2 b9 8f c7 c9 7e 0e fa 52 7b 71 e6 f8 b7 a6 ae 47 94 d7 86",
    }
    for password, hash in hashes.items():
        kcpass = mac_user._kcpassword(password)
        hash = bytes.fromhex(hash)

        # macOS adds a trailing null and pads the rest with random data
        length = len(password) + 1

        assert kcpass[:length] == hash[:length]
        assert len(kcpass) == len(hash)



def normalize_ret(ret):
    """
    Normalize the return to the format that we'll use for result checking
    """
    result = {}
    for item, descr in ret.items():
        result[item] = {
            "__run_num__": descr["__run_num__"],
            "comment": descr["comment"],
            "result": descr["result"],
            "changes": descr["changes"] != {},  # whether there where any changes
        }
    return result



"""
Simple returner for Couchbase. Optional configuration
settings are listed below, along with sane defaults.

.. code-block:: yaml

    couchbase.host:   'salt'
    couchbase.port:   8091
    couchbase.bucket: 'salt'
    couchbase.ttl: 86400
    couchbase.password: 'password'
    couchbase.skip_verify_views: False

To use the couchbase returner, append '--return couchbase' to the salt command. ex:

.. code-block:: bash

    salt '*' test.ping --return couchbase

To use the alternative configuration, append '--return_config alternative' to the salt command.

.. versionadded:: 2015.5.0

.. code-block:: bash

    salt '*' test.ping --return couchbase --return_config alternative

To override individual configuration items, append --return_kwargs '{"key:": "value"}' to the salt command.

.. versionadded:: 2016.3.0

.. code-block:: bash

    salt '*' test.ping --return couchbase --return_kwargs '{"bucket": "another-salt"}'


All of the return data will be stored in documents as follows:

JID
===
load: load obj
tgt_minions: list of minions targeted
nocache: should we not cache the return data

JID/MINION_ID
=============
return: return_data
full_ret: full load of job return
"""

import logging

import salt.utils.jid
import salt.utils.json
import salt.utils.minions

try:
    import couchbase

    HAS_DEPS = True
except ImportError:
    HAS_DEPS = False


log = logging.getLogger(__name__)


# Define the module's virtual name
__virtualname__ = "couchbase"

# some globals
COUCHBASE_CONN = None
DESIGN_NAME = "couchbase_returner"
VERIFIED_VIEWS = False

_json = salt.utils.json.import_json()


def _json_dumps(obj, **kwargs):
    return salt.utils.json.dumps(obj, _json_module=_json)


def __virtual__():
    if not HAS_DEPS:
        return False, "Could not import couchbase returner; couchbase is not installed."

    couchbase.set_json_converters(_json_dumps, salt.utils.json.loads)

    return __virtualname__


def _get_options():
    """
    Get the couchbase options from salt. Apply defaults
    if required.
    """
    return {
        "host": __opts__.get("couchbase.host", "salt"),
        "port": __opts__.get("couchbase.port", 8091),
        "bucket": __opts__.get("couchbase.bucket", "salt"),
        "password": __opts__.get("couchbase.password", ""),
    }


def _get_connection():
    """
    Global function to access the couchbase connection (and make it if its closed)
    """
    global COUCHBASE_CONN
    if COUCHBASE_CONN is None:
        opts = _get_options()
        if opts["password"]:
            COUCHBASE_CONN = couchbase.Couchbase.connect(
                host=opts["host"],
                port=opts["port"],
                bucket=opts["bucket"],
                password=opts["password"],
            )
        else:
            COUCHBASE_CONN = couchbase.Couchbase.connect(
                host=opts["host"], port=opts["port"], bucket=opts["bucket"]
            )

    return COUCHBASE_CONN


def _verify_views():
    """
    Verify that you have the views you need. This can be disabled by
    adding couchbase.skip_verify_views: True in config
    """
    global VERIFIED_VIEWS

    if VERIFIED_VIEWS or __opts__.get("couchbase.skip_verify_views", False):
        return
    cb_ = _get_connection()
    ddoc = {
        "views": {
            "jids": {
                "map": (
                    "function (doc, meta) { if (meta.id.indexOf('/') === -1 &&"
                    " doc.load){ emit(meta.id, null) } }"
                )
            },
            "jid_returns": {
                "map": (
                    "function (doc, meta) { if (meta.id.indexOf('/') > -1){ key_parts ="
                    " meta.id.split('/'); emit(key_parts[0], key_parts[1]); } }"
                )
            },
        }
    }

    try:
        curr_ddoc = cb_.design_get(DESIGN_NAME, use_devmode=False).value
        if curr_ddoc["views"] == ddoc["views"]:
            VERIFIED_VIEWS = True
            return
    except couchbase.exceptions.HTTPError:
        pass

    cb_.design_create(DESIGN_NAME, ddoc, use_devmode=False)
    VERIFIED_VIEWS = True


def _get_ttl():
    """
    Return the TTL that we should store our objects with
    """
    return __opts__.get("couchbase.ttl", 86400)


# TODO: add to returner docs-- this is a new one
def prep_jid(nocache=False, passed_jid=None):
    """
    Return a job id and prepare the job id directory
    This is the function responsible for making sure jids don't collide (unless
    its passed a jid)
    So do what you have to do to make sure that stays the case
    """
    if passed_jid is None:
        jid = salt.utils.jid.gen_jid(__opts__)
    else:
        jid = passed_jid

    cb_ = _get_connection()

    try:
        cb_.add(
            str(jid),
            {"nocache": nocache},
            ttl=_get_ttl(),
        )
    except couchbase.exceptions.KeyExistsError:
        # TODO: some sort of sleep or something? Spinning is generally bad practice
        if passed_jid is None:
            return prep_jid(nocache=nocache)

    return jid


def returner(load):
    """
    Return data to couchbase bucket
    """
    cb_ = _get_connection()

    hn_key = "{}/{}".format(load["jid"], load["id"])
    try:
        ret_doc = {"return": load["return"], "full_ret": salt.utils.json.dumps(load)}

        cb_.add(
            hn_key,
            ret_doc,
            ttl=_get_ttl(),
        )
    except couchbase.exceptions.KeyExistsError:
        log.error(
            "An extra return was detected from minion %s, please verify "
            "the minion, this could be a replay attack",
            load["id"],
        )
        return False


def save_load(jid, clear_load, minion=None):
    """
    Save the load to the specified jid
    """
    cb_ = _get_connection()

    try:
        jid_doc = cb_.get(str(jid))
    except couchbase.exceptions.NotFoundError:
        cb_.add(str(jid), {}, ttl=_get_ttl())
        jid_doc = cb_.get(str(jid))

    jid_doc.value["load"] = clear_load
    cb_.replace(str(jid), jid_doc.value, cas=jid_doc.cas, ttl=_get_ttl())

    # if you have a tgt, save that for the UI etc
    if "tgt" in clear_load and clear_load["tgt"] != "":
        ckminions = salt.utils.minions.CkMinions(__opts__)
        # Retrieve the minions list
        _res = ckminions.check_minions(
            clear_load["tgt"], clear_load.get("tgt_type", "glob")
        )
        minions = _res["minions"]
        save_minions(jid, minions)


def save_minions(jid, minions, syndic_id=None):  # pylint: disable=unused-argument
    """
    Save/update the minion list for a given jid. The syndic_id argument is
    included for API compatibility only.
    """
    cb_ = _get_connection()

    try:
        jid_doc = cb_.get(str(jid))
    except couchbase.exceptions.NotFoundError:
        log.warning("Could not write job cache file for jid: %s", jid)
        return False

    # save the minions to a cache so we can see in the UI
    if "minions" in jid_doc.value:
        jid_doc.value["minions"] = sorted(set(jid_doc.value["minions"] + minions))
    else:
        jid_doc.value["minions"] = minions
    cb_.replace(str(jid), jid_doc.value, cas=jid_doc.cas, ttl=_get_ttl())


def get_load(jid):
    """
    Return the load data that marks a specified jid
    """
    cb_ = _get_connection()

    try:
        jid_doc = cb_.get(str(jid))
    except couchbase.exceptions.NotFoundError:
        return {}

    ret = {}
    try:
        ret = jid_doc.value["load"]
        ret["Minions"] = jid_doc.value["minions"]
    except KeyError as e:
        log.error(e)

    return ret


def get_jid(jid):
    """
    Return the information returned when the specified job id was executed
    """
    cb_ = _get_connection()
    _verify_views()

    ret = {}

    for result in cb_.query(
        DESIGN_NAME, "jid_returns", key=str(jid), include_docs=True
    ):
        ret[result.value] = result.doc.value

    return ret


def get_jids():
    """
    Return a list of all job ids
    """
    cb_ = _get_connection()
    _verify_views()

    ret = {}

    for result in cb_.query(DESIGN_NAME, "jids", include_docs=True):
        ret[result.key] = _format_jid_instance(result.key, result.doc.value["load"])

    return ret


def _format_job_instance(job):
    """
    Return a properly formatted job dict
    """
    ret = {
        "Function": job.get("fun", "unknown-function"),
        "Arguments": list(job.get("arg", [])),
        # unlikely but safeguard from invalid returns
        "Target": job.get("tgt", "unknown-target"),
        "Target-type": job.get("tgt_type", "list"),
        "User": job.get("user", "root"),
    }

    if "metadata" in job:
        ret["Metadata"] = job.get("metadata", {})
    else:
        if "kwargs" in job:
            if "metadata" in job["kwargs"]:
                ret["Metadata"] = job["kwargs"].get("metadata", {})
    return ret


def _format_jid_instance(jid, job):
    """
    Return a properly formatted jid dict
    """
    ret = _format_job_instance(job)
    ret.update({"StartTime": salt.utils.jid.jid_to_time(jid)})
    return ret



#!/usr/bin/env python
#
# Copyright 2009 Facebook
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

"""A command line parsing module that lets modules define their own options.

Each module defines its own options which are added to the global
option namespace, e.g.::

    from salt.ext.tornado.options import define, options

    define("mysql_host", default="127.0.0.1:3306", help="Main user DB")
    define("memcache_hosts", default="127.0.0.1:11011", multiple=True,
           help="Main user memcache servers")

    def connect():
        db = database.Connection(options.mysql_host)
        ...

The ``main()`` method of your application does not need to be aware of all of
the options used throughout your program; they are all automatically loaded
when the modules are loaded.  However, all modules that define options
must have been imported before the command line is parsed.

Your ``main()`` method can parse the command line or parse a config file with
either::

    tornado.options.parse_command_line()
    # or
    tornado.options.parse_config_file("/etc/server.conf")

.. note:

   When using tornado.options.parse_command_line or
   tornado.options.parse_config_file, the only options that are set are
   ones that were previously defined with tornado.options.define.

Command line formats are what you would expect (``--myoption=myvalue``).
Config files are just Python files. Global names become options, e.g.::

    myoption = "myvalue"
    myotheroption = "myothervalue"

We support `datetimes <datetime.datetime>`, `timedeltas
<datetime.timedelta>`, ints, and floats (just pass a ``type`` kwarg to
`define`). We also accept multi-value options. See the documentation for
`define()` below.

`tornado.options.options` is a singleton instance of `OptionParser`, and
the top-level functions in this module (`define`, `parse_command_line`, etc)
simply call methods on it.  You may create additional `OptionParser`
instances to define isolated sets of options, such as for subcommands.

.. note::

   By default, several options are defined that will configure the
   standard `logging` module when `parse_command_line` or `parse_config_file`
   are called.  If you want Tornado to leave the logging configuration
   alone so you can manage it yourself, either pass ``--logging=none``
   on the command line or do the following to disable it in code::

       from salt.ext.tornado.options import options, parse_command_line
       options.logging = None
       parse_command_line()

.. versionchanged:: 4.3
   Dashes and underscores are fully interchangeable in option names;
   options can be defined, set, and read with any mix of the two.
   Dashes are typical for command-line usage while config files require
   underscores.
"""
# pylint: skip-file

from __future__ import absolute_import, division, print_function

import datetime
import numbers
import re
import sys
import os
import textwrap

from salt.ext.tornado.escape import _unicode, native_str
from salt.ext.tornado.log import define_logging_options
from salt.ext.tornado import stack_context
from salt.ext.tornado.util import basestring_type, exec_in


class Error(Exception):
    """Exception raised by errors in the options module."""
    pass


class OptionParser(object):
    """A collection of options, a dictionary with object-like access.

    Normally accessed via static functions in the `tornado.options` module,
    which reference a global instance.
    """
    def __init__(self):
        # we have to use self.__dict__ because we override setattr.
        self.__dict__['_options'] = {}
        self.__dict__['_parse_callbacks'] = []
        self.define("help", type=bool, help="show this help information",
                    callback=self._help_callback)

    def _normalize_name(self, name):
        return name.replace('_', '-')

    def __getattr__(self, name):
        name = self._normalize_name(name)
        if isinstance(self._options.get(name), _Option):
            return self._options[name].value()
        raise AttributeError("Unrecognized option %r" % name)

    def __setattr__(self, name, value):
        name = self._normalize_name(name)
        if isinstance(self._options.get(name), _Option):
            return self._options[name].set(value)
        raise AttributeError("Unrecognized option %r" % name)

    def __iter__(self):
        return (opt.name for opt in self._options.values())

    def __contains__(self, name):
        name = self._normalize_name(name)
        return name in self._options

    def __getitem__(self, name):
        return self.__getattr__(name)

    def __setitem__(self, name, value):
        return self.__setattr__(name, value)

    def items(self):
        """A sequence of (name, value) pairs.

        .. versionadded:: 3.1
        """
        return [(opt.name, opt.value()) for name, opt in self._options.items()]

    def groups(self):
        """The set of option-groups created by ``define``.

        .. versionadded:: 3.1
        """
        return set(opt.group_name for opt in self._options.values())

    def group_dict(self, group):
        """The names and values of options in a group.

        Useful for copying options into Application settings::

            from salt.ext.tornado.options import define, parse_command_line, options

            define('template_path', group='application')
            define('static_path', group='application')

            parse_command_line()

            application = Application(
                handlers, **options.group_dict('application'))

        .. versionadded:: 3.1
        """
        return dict(
            (opt.name, opt.value()) for name, opt in self._options.items()
            if not group or group == opt.group_name)

    def as_dict(self):
        """The names and values of all options.

        .. versionadded:: 3.1
        """
        return dict(
            (opt.name, opt.value()) for name, opt in self._options.items())

    def define(self, name, default=None, type=None, help=None, metavar=None,
               multiple=False, group=None, callback=None):
        """Defines a new command line option.

        If ``type`` is given (one of str, float, int, datetime, or timedelta)
        or can be inferred from the ``default``, we parse the command line
        arguments based on the given type. If ``multiple`` is True, we accept
        comma-separated values, and the option value is always a list.

        For multi-value integers, we also accept the syntax ``x:y``, which
        turns into ``range(x, y)`` - very useful for long integer ranges.

        ``help`` and ``metavar`` are used to construct the
        automatically generated command line help string. The help
        message is formatted like::

           --name=METAVAR      help string

        ``group`` is used to group the defined options in logical
        groups. By default, command line options are grouped by the
        file in which they are defined.

        Command line option names must be unique globally. They can be parsed
        from the command line with `parse_command_line` or parsed from a
        config file with `parse_config_file`.

        If a ``callback`` is given, it will be run with the new value whenever
        the option is changed.  This can be used to combine command-line
        and file-based options::

            define("config", type=str, help="path to config file",
                   callback=lambda path: parse_config_file(path, final=False))

        With this definition, options in the file specified by ``--config`` will
        override options set earlier on the command line, but can be overridden
        by later flags.
        """
        normalized = self._normalize_name(name)
        if normalized in self._options:
            raise Error("Option %r already defined in %s" %
                        (normalized, self._options[normalized].file_name))
        frame = sys._getframe(0)
        options_file = frame.f_code.co_filename

        # Can be called directly, or through top level define() fn, in which
        # case, step up above that frame to look for real caller.
        if (frame.f_back.f_code.co_filename == options_file and
                frame.f_back.f_code.co_name == 'define'):
            frame = frame.f_back

        file_name = frame.f_back.f_code.co_filename
        if file_name == options_file:
            file_name = ""
        if type is None:
            if not multiple and default is not None:
                type = default.__class__
            else:
                type = str
        if group:
            group_name = group
        else:
            group_name = file_name
        option = _Option(name, file_name=file_name,
                         default=default, type=type, help=help,
                         metavar=metavar, multiple=multiple,
                         group_name=group_name,
                         callback=callback)
        self._options[normalized] = option

    def parse_command_line(self, args=None, final=True):
        """Parses all options given on the command line (defaults to
        `sys.argv`).

        Note that ``args[0]`` is ignored since it is the program name
        in `sys.argv`.

        We return a list of all arguments that are not parsed as options.

        If ``final`` is ``False``, parse callbacks will not be run.
        This is useful for applications that wish to combine configurations
        from multiple sources.
        """
        if args is None:
            args = sys.argv
        remaining = []
        for i in range(1, len(args)):
            # All things after the last option are command line arguments
            if not args[i].startswith("-"):
                remaining = args[i:]
                break
            if args[i] == "--":
                remaining = args[i + 1:]
                break
            arg = args[i].lstrip("-")
            name, equals, value = arg.partition("=")
            name = self._normalize_name(name)
            if name not in self._options:
                self.print_help()
                raise Error('Unrecognized command line option: %r' % name)
            option = self._options[name]
            if not equals:
                if option.type == bool:
                    value = "true"
                else:
                    raise Error('Option %r requires a value' % name)
            option.parse(value)

        if final:
            self.run_parse_callbacks()

        return remaining

    def parse_config_file(self, path, final=True):
        """Parses and loads the Python config file at the given path.

        If ``final`` is ``False``, parse callbacks will not be run.
        This is useful for applications that wish to combine configurations
        from multiple sources.

        .. versionchanged:: 4.1
           Config files are now always interpreted as utf-8 instead of
           the system default encoding.

        .. versionchanged:: 4.4
           The special variable ``__file__`` is available inside config
           files, specifying the absolute path to the config file itself.
        """
        config = {'__file__': os.path.abspath(path)}
        with open(path, 'rb') as f:
            exec_in(native_str(f.read()), config, config)
        for name in config:
            normalized = self._normalize_name(name)
            if normalized in self._options:
                self._options[normalized].set(config[name])

        if final:
            self.run_parse_callbacks()

    def print_help(self, file=None):
        """Prints all the command line options to stderr (or another file)."""
        if file is None:
            file = sys.stderr
        print("Usage: %s [OPTIONS]" % sys.argv[0], file=file)
        print("\nOptions:\n", file=file)
        by_group = {}
        for option in self._options.values():
            by_group.setdefault(option.group_name, []).append(option)

        for filename, o in sorted(by_group.items()):
            if filename:
                print("\n%s options:\n" % os.path.normpath(filename), file=file)
            o.sort(key=lambda option: option.name)
            for option in o:
                # Always print names with dashes in a CLI context.
                prefix = self._normalize_name(option.name)
                if option.metavar:
                    prefix += "=" + option.metavar
                description = option.help or ""
                if option.default is not None and option.default != '':
                    description += " (default %s)" % option.default
                lines = textwrap.wrap(description, 79 - 35)
                if len(prefix) > 30 or len(lines) == 0:
                    lines.insert(0, '')
                print("  --%-30s %s" % (prefix, lines[0]), file=file)
                for line in lines[1:]:
                    print("%-34s %s" % (' ', line), file=file)
        print(file=file)

    def _help_callback(self, value):
        if value:
            self.print_help()
            sys.exit(0)

    def add_parse_callback(self, callback):
        """Adds a parse callback, to be invoked when option parsing is done."""
        self._parse_callbacks.append(stack_context.wrap(callback))

    def run_parse_callbacks(self):
        for callback in self._parse_callbacks:
            callback()

    def mockable(self):
        """Returns a wrapper around self that is compatible with
        `mock.patch <unittest.mock.patch>`.

        The `mock.patch <unittest.mock.patch>` function (included in
        the standard library `unittest.mock` package since Python 3.3,
        or in the third-party ``mock`` package for older versions of
        Python) is incompatible with objects like ``options`` that
        override ``__getattr__`` and ``__setattr__``.  This function
        returns an object that can be used with `mock.patch.object
        <unittest.mock.patch.object>` to modify option values::

            with mock.patch.object(options.mockable(), 'name', value):
                assert options.name == value
        """
        return _Mockable(self)


class _Mockable(object):
    """`mock.patch` compatible wrapper for `OptionParser`.

    As of ``mock`` version 1.0.1, when an object uses ``__getattr__``
    hooks instead of ``__dict__``, ``patch.__exit__`` tries to delete
    the attribute it set instead of setting a new one (assuming that
    the object does not catpure ``__setattr__``, so the patch
    created a new attribute in ``__dict__``).

    _Mockable's getattr and setattr pass through to the underlying
    OptionParser, and delattr undoes the effect of a previous setattr.
    """
    def __init__(self, options):
        # Modify __dict__ directly to bypass __setattr__
        self.__dict__['_options'] = options
        self.__dict__['_originals'] = {}

    def __getattr__(self, name):
        return getattr(self._options, name)

    def __setattr__(self, name, value):
        assert name not in self._originals, "don't reuse mockable objects"
        self._originals[name] = getattr(self._options, name)
        setattr(self._options, name, value)

    def __delattr__(self, name):
        setattr(self._options, name, self._originals.pop(name))


class _Option(object):
    UNSET = object()

    def __init__(self, name, default=None, type=basestring_type, help=None,
                 metavar=None, multiple=False, file_name=None, group_name=None,
                 callback=None):
        if default is None and multiple:
            default = []
        self.name = name
        self.type = type
        self.help = help
        self.metavar = metavar
        self.multiple = multiple
        self.file_name = file_name
        self.group_name = group_name
        self.callback = callback
        self.default = default
        self._value = _Option.UNSET

    def value(self):
        return self.default if self._value is _Option.UNSET else self._value

    def parse(self, value):
        _parse = {
            datetime.datetime: self._parse_datetime,
            datetime.timedelta: self._parse_timedelta,
            bool: self._parse_bool,
            basestring_type: self._parse_string,
        }.get(self.type, self.type)
        if self.multiple:
            self._value = []
            for part in value.split(","):
                if issubclass(self.type, numbers.Integral):
                    # allow ranges of the form X:Y (inclusive at both ends)
                    lo, _, hi = part.partition(":")
                    lo = _parse(lo)
                    hi = _parse(hi) if hi else lo
                    self._value.extend(range(lo, hi + 1))
                else:
                    self._value.append(_parse(part))
        else:
            self._value = _parse(value)
        if self.callback is not None:
            self.callback(self._value)
        return self.value()

    def set(self, value):
        if self.multiple:
            if not isinstance(value, list):
                raise Error("Option %r is required to be a list of %s" %
                            (self.name, self.type.__name__))
            for item in value:
                if item is not None and not isinstance(item, self.type):
                    raise Error("Option %r is required to be a list of %s" %
                                (self.name, self.type.__name__))
        else:
            if value is not None and not isinstance(value, self.type):
                raise Error("Option %r is required to be a %s (%s given)" %
                            (self.name, self.type.__name__, type(value)))
        self._value = value
        if self.callback is not None:
            self.callback(self._value)

    # Supported date/time formats in our options
    _DATETIME_FORMATS = [
        "%a %b %d %H:%M:%S %Y",
        "%Y-%m-%d %H:%M:%S",
        "%Y-%m-%d %H:%M",
        "%Y-%m-%dT%H:%M",
        "%Y%m%d %H:%M:%S",
        "%Y%m%d %H:%M",
        "%Y-%m-%d",
        "%Y%m%d",
        "%H:%M:%S",
        "%H:%M",
    ]

    def _parse_datetime(self, value):
        for format in self._DATETIME_FORMATS:
            try:
                return datetime.datetime.strptime(value, format)
            except ValueError:
                pass
        raise Error('Unrecognized date/time format: %r' % value)

    _TIMEDELTA_ABBREV_DICT = {
        'h': 'hours',
        'm': 'minutes',
        'min': 'minutes',
        's': 'seconds',
        'sec': 'seconds',
        'ms': 'milliseconds',
        'us': 'microseconds',
        'd': 'days',
        'w': 'weeks',
    }

    _FLOAT_PATTERN = r'[-+]?(?:\d+(?:\.\d*)?|\.\d+)(?:[eE][-+]?\d+)?'

    _TIMEDELTA_PATTERN = re.compile(
        r'\s*(%s)\s*(\w*)\s*' % _FLOAT_PATTERN, re.IGNORECASE)

    def _parse_timedelta(self, value):
        try:
            sum = datetime.timedelta()
            start = 0
            while start < len(value):
                m = self._TIMEDELTA_PATTERN.match(value, start)
                if not m:
                    raise Exception()
                num = float(m.group(1))
                units = m.group(2) or 'seconds'
                units = self._TIMEDELTA_ABBREV_DICT.get(units, units)
                sum += datetime.timedelta(**{units: num})
                start = m.end()
            return sum
        except Exception:
            raise

    def _parse_bool(self, value):
        return value.lower() not in ("false", "0", "f")

    def _parse_string(self, value):
        return _unicode(value)


options = OptionParser()
"""Global options object.

All defined options are available as attributes on this object.
"""


def define(name, default=None, type=None, help=None, metavar=None,
           multiple=False, group=None, callback=None):
    """Defines an option in the global namespace.

    See `OptionParser.define`.
    """
    return options.define(name, default=default, type=type, help=help,
                          metavar=metavar, multiple=multiple, group=group,
                          callback=callback)


def parse_command_line(args=None, final=True):
    """Parses global options from the command line.

    See `OptionParser.parse_command_line`.
    """
    return options.parse_command_line(args, final=final)


def parse_config_file(path, final=True):
    """Parses global options from a config file.

    See `OptionParser.parse_config_file`.
    """
    return options.parse_config_file(path, final=final)


def print_help(file=None):
    """Prints all the command line options to stderr (or another file).

    See `OptionParser.print_help`.
    """
    return options.print_help(file)


def add_parse_callback(callback):
    """Adds a parse callback, to be invoked when option parsing is done.

    See `OptionParser.add_parse_callback`
    """
    options.add_parse_callback(callback)


# Default options
define_logging_options(options)



"""
Run processes as a different user in Windows
"""

# Import Python Libraries
import ctypes
import logging
import os
import subprocess
import time

from salt.exceptions import CommandExecutionError

try:
    import psutil

    HAS_PSUTIL = True
except ImportError:
    HAS_PSUTIL = False

try:
    import msvcrt

    import pywintypes
    import win32api
    import win32con
    import win32event
    import win32pipe
    import win32process
    import win32profile
    import win32security

    import salt.platform.win

    HAS_WIN32 = True
except ImportError:
    HAS_WIN32 = False


log = logging.getLogger(__name__)


# Although utils are often directly imported, it is also possible to use the
# loader.
def __virtual__():
    """
    Only load if Win32 Libraries are installed
    """
    if not HAS_WIN32 or not HAS_PSUTIL:
        return False, "This utility requires pywin32 and psutil"

    return "win_runas"


def split_username(username):
    """
    Splits out the username from the domain name and returns both.
    """
    domain = "."
    user_name = str(username)
    # Domain users with User Principal Name (UPN): user@DOMAIN
    if "@" in user_name:
        user_name, domain = user_name.split("@", maxsplit=1)
        domain = domain.removesuffix(".local")
    # Domain users with Down-Level Logon Name: DOMAIN\user
    if "\\" in user_name:
        domain, user_name = user_name.split("\\", maxsplit=1)
    return str(user_name), str(domain)


def create_env(user_token, inherit, timeout=1):
    """
    CreateEnvironmentBlock might fail when we close a login session and then
    try to re-open one very quickly. Run the method multiple times to work
    around the async nature of logoffs.
    """
    start = time.time()
    env = None
    exc = None
    while True:
        try:
            env = win32profile.CreateEnvironmentBlock(user_token, False)
        except pywintypes.error as exc:
            pass
        else:
            break
        if time.time() - start > timeout:
            break
    if env is not None:
        return env
    if exc is not None:
        raise exc


def runas(cmd, username, password=None, cwd=None):
    """
    Run a command as another user. If the process is running as an admin or
    system account this method does not require a password. Other non
    privileged accounts need to provide a password for the user to runas.
    Commands are run in with the highest level privileges possible for the
    account provided.
    """
    # Sometimes this comes in as an int. LookupAccountName can't handle an int
    # Let's make it a string if it's anything other than a string
    if not isinstance(username, str):
        username = str(username)
    # Validate the domain and sid exist for the username
    try:
        _, domain, _ = win32security.LookupAccountName(None, username)
        username, _ = split_username(username)
    except pywintypes.error as exc:
        message = win32api.FormatMessage(exc.winerror).rstrip("\n")
        raise CommandExecutionError(message)

    # Elevate the token from the current process
    access = win32security.TOKEN_QUERY | win32security.TOKEN_ADJUST_PRIVILEGES
    th = win32security.OpenProcessToken(win32api.GetCurrentProcess(), access)
    salt.platform.win.elevate_token(th)

    # Try to impersonate the SYSTEM user. This process needs to be running as a
    # user who as been granted the SeImpersonatePrivilege, Administrator
    # accounts have this permission by default.
    try:
        impersonation_token = salt.platform.win.impersonate_sid(
            salt.platform.win.SYSTEM_SID,
            session_id=0,
            privs=["SeTcbPrivilege"],
        )
    except OSError:
        log.debug("Unable to impersonate SYSTEM user")
        impersonation_token = None
        win32api.CloseHandle(th)

    if isinstance(cmd, (list, tuple)):
        # CreateProcess parameter lpCommandLine must be a string.
        # Since it is called directly and not via the subprocess module,
        # the arguments must be processed manually.
        cmd = subprocess.list2cmdline(cmd)

    # Impersonation of the SYSTEM user failed. Fallback to an un-privileged
    # runas.
    if not impersonation_token:
        log.debug("No impersonation token, using unprivileged runas")
        return runas_unpriv(cmd, username, password, cwd)

    if domain == "NT AUTHORITY":
        # Logon as a system level account, SYSTEM, LOCAL SERVICE, or NETWORK
        # SERVICE.
        user_token = win32security.LogonUser(
            username,
            domain,
            "",
            win32con.LOGON32_LOGON_SERVICE,
            win32con.LOGON32_PROVIDER_DEFAULT,
        )
    elif password:
        # Login with a password.
        user_token = win32security.LogonUser(
            username,
            domain,
            password,
            win32con.LOGON32_LOGON_INTERACTIVE,
            win32con.LOGON32_PROVIDER_DEFAULT,
        )
    else:
        # Login without a password. This always returns an elevated token.
        user_token = salt.platform.win.logon_msv1_s4u(username).Token

    # Get a linked user token to elevate if needed
    elevation_type = win32security.GetTokenInformation(
        user_token, win32security.TokenElevationType
    )
    if elevation_type > 1:
        user_token = win32security.GetTokenInformation(
            user_token, win32security.TokenLinkedToken
        )

    # Elevate the user token
    salt.platform.win.elevate_token(user_token)

    # Make sure the user's token has access to a windows station and desktop
    salt.platform.win.grant_winsta_and_desktop(user_token)

    # Create pipes for standard in, out and error streams
    security_attributes = win32security.SECURITY_ATTRIBUTES()
    security_attributes.bInheritHandle = 1

    stdin_read, stdin_write = win32pipe.CreatePipe(security_attributes, 0)
    stdin_read = salt.platform.win.make_inheritable(stdin_read)

    stdout_read, stdout_write = win32pipe.CreatePipe(security_attributes, 0)
    stdout_write = salt.platform.win.make_inheritable(stdout_write)

    stderr_read, stderr_write = win32pipe.CreatePipe(security_attributes, 0)
    stderr_write = salt.platform.win.make_inheritable(stderr_write)

    # Run the process without showing a window.
    creationflags = (
        win32process.CREATE_NO_WINDOW
        | win32process.CREATE_NEW_CONSOLE
        | win32process.CREATE_SUSPENDED
    )

    flags = win32con.STARTF_USESTDHANDLES
    flags |= win32con.STARTF_USESHOWWINDOW
    startup_info = salt.platform.win.STARTUPINFO(
        dwFlags=flags,
        hStdInput=stdin_read.handle,
        hStdOutput=stdout_write.handle,
        hStdError=stderr_write.handle,
    )

    # Create the environment for the user
    env = create_env(user_token, False)

    hProcess = None
    try:
        # Start the process in a suspended state.
        process_info = salt.platform.win.CreateProcessWithTokenW(
            int(user_token),
            logonflags=1,
            applicationname=None,
            commandline=cmd,
            currentdirectory=cwd,
            creationflags=creationflags,
            startupinfo=startup_info,
            environment=env,
        )

        hProcess = process_info.hProcess
        hThread = process_info.hThread
        dwProcessId = process_info.dwProcessId
        dwThreadId = process_info.dwThreadId

        # We don't use these so let's close the handle
        salt.platform.win.kernel32.CloseHandle(stdin_write.handle)
        salt.platform.win.kernel32.CloseHandle(stdout_write.handle)
        salt.platform.win.kernel32.CloseHandle(stderr_write.handle)

        ret = {"pid": dwProcessId}
        # Resume the process
        psutil.Process(dwProcessId).resume()

        # Wait for the process to exit and get its return code.
        if (
            win32event.WaitForSingleObject(hProcess, win32event.INFINITE)
            == win32con.WAIT_OBJECT_0
        ):
            exitcode = win32process.GetExitCodeProcess(hProcess)
            ret["retcode"] = exitcode

        # Read standard out
        fd_out = msvcrt.open_osfhandle(stdout_read.handle, os.O_RDONLY | os.O_TEXT)
        with os.fdopen(fd_out, "r") as f_out:
            stdout = f_out.read()
            ret["stdout"] = stdout.strip()

        # Read standard error
        fd_err = msvcrt.open_osfhandle(stderr_read.handle, os.O_RDONLY | os.O_TEXT)
        with os.fdopen(fd_err, "r") as f_err:
            stderr = f_err.read()
            ret["stderr"] = stderr
    finally:
        if hProcess is not None:
            salt.platform.win.kernel32.CloseHandle(hProcess)
        win32api.CloseHandle(th)
        win32api.CloseHandle(user_token)
        if impersonation_token:
            win32security.RevertToSelf()
        win32api.CloseHandle(impersonation_token)

    return ret


def runas_unpriv(cmd, username, password, cwd=None):
    """
    Runas that works for non-privileged users
    """
    # Sometimes this comes in as an int. LookupAccountName can't handle an int
    # Let's make it a string if it's anything other than a string
    if not isinstance(username, str):
        username = str(username)
    # Validate the domain and sid exist for the username
    try:
        _, domain, _ = win32security.LookupAccountName(None, username)
        username, _ = split_username(username)
    except pywintypes.error as exc:
        message = win32api.FormatMessage(exc.winerror).rstrip("\n")
        raise CommandExecutionError(message)

    # Create a pipe to set as stdout in the child. The write handle needs to be
    # inheritable.
    c2pread, c2pwrite = salt.platform.win.CreatePipe(
        inherit_read=False,
        inherit_write=True,
    )
    errread, errwrite = salt.platform.win.CreatePipe(
        inherit_read=False,
        inherit_write=True,
    )

    # Create inheritable copy of the stdin
    stdin = salt.platform.win.kernel32.GetStdHandle(
        salt.platform.win.STD_INPUT_HANDLE,
    )
    dupin = salt.platform.win.DuplicateHandle(srchandle=stdin, inherit=True)

    # Get startup info structure
    flags = win32con.STARTF_USESTDHANDLES
    flags |= win32con.STARTF_USESHOWWINDOW
    startup_info = salt.platform.win.STARTUPINFO(
        dwFlags=flags,
        hStdInput=dupin,
        hStdOutput=c2pwrite,
        hStdError=errwrite,
    )

    try:
        # Run command and return process info structure
        process_info = salt.platform.win.CreateProcessWithLogonW(
            username=username,
            domain=domain,
            password=password,
            logonflags=salt.platform.win.LOGON_WITH_PROFILE,
            commandline=cmd,
            startupinfo=startup_info,
            currentdirectory=cwd,
        )
        salt.platform.win.kernel32.CloseHandle(process_info.hThread)
    finally:
        salt.platform.win.kernel32.CloseHandle(dupin)
        salt.platform.win.kernel32.CloseHandle(c2pwrite)
        salt.platform.win.kernel32.CloseHandle(errwrite)

    # Initialize ret and set first element
    ret = {"pid": process_info.dwProcessId}

    # Get Standard Out
    fd_out = msvcrt.open_osfhandle(c2pread, os.O_RDONLY | os.O_TEXT)
    with os.fdopen(fd_out, "r") as f_out:
        ret["stdout"] = f_out.read()

    # Get Standard Error
    fd_err = msvcrt.open_osfhandle(errread, os.O_RDONLY | os.O_TEXT)
    with os.fdopen(fd_err, "r") as f_err:
        ret["stderr"] = f_err.read()

    # Get Return Code
    if (
        salt.platform.win.kernel32.WaitForSingleObject(
            process_info.hProcess, win32event.INFINITE
        )
        == win32con.WAIT_OBJECT_0
    ):
        exitcode = salt.platform.win.wintypes.DWORD()
        salt.platform.win.kernel32.GetExitCodeProcess(
            process_info.hProcess, ctypes.byref(exitcode)
        )
        ret["retcode"] = exitcode.value

    # Close handle to process
    salt.platform.win.kernel32.CloseHandle(process_info.hProcess)

    return ret



"""
Work with virtual machines managed by Vagrant.

.. versionadded:: 2018.3.0

Mapping between a Salt node id and the Vagrant machine name
(and the path to the Vagrantfile where it is defined)
is stored in a Salt sdb database on the Vagrant host (minion) machine.
In order to use this module, sdb must be configured. An SQLite
database is the recommended storage method.  The URI used for
the sdb lookup is "sdb://vagrant_sdb_data".

requirements:
   - the VM host machine must have salt-minion, Vagrant and a vm provider installed.
   - the VM host must have a valid definition for `sdb://vagrant_sdb_data`

    Configuration example:

    .. code-block:: yaml

        # file /etc/salt/minion.d/vagrant_sdb.conf
        vagrant_sdb_data:
          driver: sqlite3
          database: /var/cache/salt/vagrant.sqlite
          table: sdb
          create_table: True

"""

import logging
import os

import salt.utils.files
import salt.utils.path
import salt.utils.stringutils
from salt._compat import ipaddress
from salt.exceptions import CommandExecutionError, SaltInvocationError

log = logging.getLogger(__name__)

__virtualname__ = "vagrant"

VAGRANT_SDB_URL = "sdb://vagrant_sdb_data/"


def __virtual__():
    """
    run Vagrant commands if possible
    """
    if salt.utils.path.which("vagrant") is None:
        return (
            False,
            "The vagrant module could not be loaded: vagrant command not found",
        )
    return __virtualname__


def _build_sdb_uri(key):
    """
    returns string used to fetch data for "key" from the sdb store.

    Salt node id's are used as the key for vm_ dicts.

    """
    return f"{VAGRANT_SDB_URL}{key}"


def _build_machine_uri(machine, cwd):
    """
    returns string used to fetch id names from the sdb store.

    the cwd and machine name are concatenated with '?' which should
    never collide with a Salt node id -- which is important since we
    will be storing both in the same table.
    """
    key = f"{machine}?{os.path.abspath(cwd)}"
    return _build_sdb_uri(key)


def _update_vm_info(name, vm_):
    """store the vm_ information keyed by name"""
    __utils__["sdb.sdb_set"](_build_sdb_uri(name), vm_, __opts__)

    # store machine-to-name mapping, too
    if vm_["machine"]:
        __utils__["sdb.sdb_set"](
            _build_machine_uri(vm_["machine"], vm_.get("cwd", ".")), name, __opts__
        )


def get_vm_info(name):
    """
    get the information for a VM.

    :param name: salt_id name
    :return: dictionary of {'machine': x, 'cwd': y, ...}.
    """
    try:
        vm_ = __utils__["sdb.sdb_get"](_build_sdb_uri(name), __opts__)
    except KeyError:
        raise SaltInvocationError(
            "Probable sdb driver not found. Check your configuration."
        )
    if vm_ is None or "machine" not in vm_:
        raise SaltInvocationError(f"No Vagrant machine defined for Salt_id {name}")
    return vm_


def get_machine_id(machine, cwd):
    """
    returns the salt_id name of the Vagrant VM

    :param machine: the Vagrant machine name
    :param cwd: the path to Vagrantfile
    :return: salt_id name
    """
    name = __utils__["sdb.sdb_get"](_build_machine_uri(machine, cwd), __opts__)
    return name


def _erase_vm_info(name):
    """
    erase the information for a VM the we are destroying.

    some sdb drivers (such as the SQLite driver we expect to use)
    do not have a `delete` method, so if the delete fails, we have
    to replace the with a blank entry.
    """
    try:
        # delete the machine record
        vm_ = get_vm_info(name)
        if vm_["machine"]:
            key = _build_machine_uri(vm_["machine"], vm_.get("cwd", "."))
            try:
                __utils__["sdb.sdb_delete"](key, __opts__)
            except KeyError:
                # no delete method found -- load a blank value
                __utils__["sdb.sdb_set"](key, None, __opts__)
    except Exception:  # pylint: disable=broad-except
        pass

    uri = _build_sdb_uri(name)
    try:
        # delete the name record
        __utils__["sdb.sdb_delete"](uri, __opts__)
    except KeyError:
        # no delete method found -- load an empty dictionary
        __utils__["sdb.sdb_set"](uri, {}, __opts__)
    except Exception:  # pylint: disable=broad-except
        pass


def _vagrant_ssh_config(vm_):
    """
    get the information for ssh communication from the new VM

    :param vm_: the VM's info as we have it now
    :return: dictionary of ssh stuff
    """
    machine = vm_["machine"]
    log.info("requesting vagrant ssh-config for VM %s", machine or "(default)")
    cmd = f"vagrant ssh-config {machine}"
    reply = __salt__["cmd.shell"](
        cmd, runas=vm_.get("runas"), cwd=vm_.get("cwd"), ignore_retcode=True
    )
    ssh_config = {}
    for line in reply.split("\n"):  # build a dictionary of the text reply
        tokens = line.strip().split()
        if len(tokens) == 2:  # each two-token line becomes a key:value pair
            ssh_config[tokens[0]] = tokens[1]
    log.debug("ssh_config=%s", repr(ssh_config))
    return ssh_config


def version():
    """
    Return the version of Vagrant on the minion

    CLI Example:

    .. code-block:: bash

        salt '*' vagrant.version
    """
    cmd = "vagrant -v"
    return __salt__["cmd.shell"](cmd)


def list_domains():
    """
    Return a list of the salt_id names of all available Vagrant VMs on
    this host without regard to the path where they are defined.

    CLI Example:

    .. code-block:: bash

        salt '*' vagrant.list_domains --log-level=info

    The log shows information about all known Vagrant environments
    on this machine. This data is cached and may not be completely
    up-to-date.
    """
    vms = []
    cmd = "vagrant global-status"
    reply = __salt__["cmd.shell"](cmd)
    log.debug("--->\n%s", reply)
    for line in reply.split("\n"):  # build a list of the text reply
        tokens = line.strip().split()
        try:
            _ = int(tokens[0], 16)  # valid id numbers are hexadecimal
        except (ValueError, IndexError):
            continue  # skip lines without valid id numbers
        machine = tokens[1]
        cwd = tokens[-1]
        name = get_machine_id(machine, cwd)
        if name:
            vms.append(name)
    return vms


def list_active_vms(cwd=None):
    """
    Return a list of machine names for active virtual machine on the host,
    which are defined in the Vagrantfile at the indicated path.

    CLI Example:

    .. code-block:: bash

        salt '*' vagrant.list_active_vms  cwd=/projects/project_1
    """
    vms = []
    cmd = "vagrant status"
    reply = __salt__["cmd.shell"](cmd, cwd=cwd)
    log.info("--->\n%s", reply)
    for line in reply.split("\n"):  # build a list of the text reply
        tokens = line.strip().split()
        if len(tokens) > 1:
            if tokens[1] == "running":
                vms.append(tokens[0])
    return vms


def list_inactive_vms(cwd=None):
    """
    Return a list of machine names for inactive virtual machine on the host,
    which are defined in the Vagrantfile at the indicated path.

    CLI Example:

    .. code-block:: bash

        salt '*' virt.list_inactive_vms cwd=/projects/project_1
    """
    vms = []
    cmd = "vagrant status"
    reply = __salt__["cmd.shell"](cmd, cwd=cwd)
    log.info("--->\n%s", reply)
    for line in reply.split("\n"):  # build a list of the text reply
        tokens = line.strip().split()
        if len(tokens) > 1 and tokens[-1].endswith(")"):
            if tokens[1] != "running":
                vms.append(tokens[0])
    return vms


def vm_state(name="", cwd=None):
    """
    Return list of information for all the vms indicating their state.

    If you pass a VM name in as an argument then it will return info
    for just the named VM, otherwise it will return all VMs defined by
    the Vagrantfile in the `cwd` directory.

    CLI Example:

    .. code-block:: bash

        salt '*' vagrant.vm_state <name>  cwd=/projects/project_1

    returns a list of dictionaries with machine name, state, provider,
    and salt_id name.

    .. code-block:: python

        datum = {'machine': _, # Vagrant machine name,
                 'state': _, # string indicating machine state, like 'running'
                 'provider': _, # the Vagrant VM provider
                 'name': _} # salt_id name

    Known bug: if there are multiple machines in your Vagrantfile, and you request
    the status of the ``primary`` machine, which you defined by leaving the ``machine``
    parameter blank, then you may receive the status of all of them.
    Please specify the actual machine name for each VM if there are more than one.

    """

    if name:
        vm_ = get_vm_info(name)
        machine = vm_["machine"]
        cwd = vm_["cwd"] or cwd  # usually ignore passed-in cwd
    else:
        if not cwd:
            raise SaltInvocationError(
                f"Path to Vagranfile must be defined, but cwd={cwd}"
            )
        machine = ""

    info = []
    cmd = f"vagrant status {machine}"
    reply = __salt__["cmd.shell"](cmd, cwd)
    log.info("--->\n%s", reply)
    for line in reply.split("\n"):  # build a list of the text reply
        tokens = line.strip().split()
        if len(tokens) > 1 and tokens[-1].endswith(")"):
            try:
                datum = {
                    "machine": tokens[0],
                    "state": " ".join(tokens[1:-1]),
                    "provider": tokens[-1].lstrip("(").rstrip(")"),
                    "name": get_machine_id(tokens[0], cwd),
                }
                info.append(datum)
            except IndexError:
                pass
    return info


def init(
    name,  # Salt_id for created VM
    cwd=None,  # path to find Vagrantfile
    machine="",  # name of machine in Vagrantfile
    runas=None,  # username who owns Vagrant box
    start=False,  # start the machine when initialized
    vagrant_provider="",  # vagrant provider (default=virtualbox)
    vm=None,  # a dictionary of VM configuration settings
):
    """
    Initialize a new Vagrant VM.

    This inputs all the information needed to start a Vagrant VM.  These settings are stored in
    a Salt sdb database on the Vagrant host minion and used to start, control, and query the
    guest VMs. The salt_id assigned here is the key field for that database and must be unique.

    :param name: The salt_id name you will use to control this VM
    :param cwd: The path to the directory where the Vagrantfile is located
    :param machine: The machine name in the Vagrantfile. If blank, the primary machine will be used.
    :param runas: The username on the host who owns the Vagrant work files.
    :param start: (default: False) Start the virtual machine now.
    :param vagrant_provider: The name of a Vagrant VM provider (if not the default).
    :param vm: Optionally, all the above information may be supplied in this dictionary.
    :return: A string indicating success, or False.

    CLI Example:

    .. code-block:: bash

        salt <host> vagrant.init <salt_id> /path/to/Vagrantfile
        salt my_laptop vagrant.init x1 /projects/bevy_master machine=quail1
    """
    vm_ = {} if vm is None else vm.copy()  # passed configuration data
    vm_["name"] = name
    # passed-in keyword arguments overwrite vm dictionary values
    vm_["cwd"] = cwd or vm_.get("cwd")
    if not vm_["cwd"]:
        raise SaltInvocationError(
            'Path to Vagrantfile must be defined by "cwd" argument'
        )
    vm_["machine"] = machine or vm_.get("machine", machine)
    vm_["runas"] = runas or vm_.get("runas", runas)
    vm_["vagrant_provider"] = vagrant_provider or vm_.get("vagrant_provider", "")
    _update_vm_info(name, vm_)

    if start:
        log.debug("Starting VM %s", name)
        ret = _start(name, vm_)
    else:
        ret = "Name {} defined using VM {}".format(name, vm_["machine"] or "(default)")
    return ret


def start(name):
    """
    Start (vagrant up) a virtual machine defined by salt_id name.
    The machine must have been previously defined using "vagrant.init".

    CLI Example:

    .. code-block:: bash

        salt <host> vagrant.start <salt_id>
    """
    vm_ = get_vm_info(name)
    return _start(name, vm_)


def _start(
    name, vm_
):  # internal call name, because "start" is a keyword argument to vagrant.init

    try:
        machine = vm_["machine"]
    except KeyError:
        raise SaltInvocationError(f"No Vagrant machine defined for Salt_id {name}")

    vagrant_provider = vm_.get("vagrant_provider", "")
    provider_ = f"--provider={vagrant_provider}" if vagrant_provider else ""
    cmd = f"vagrant up {machine} {provider_}"
    ret = __salt__["cmd.run_all"](
        cmd, runas=vm_.get("runas"), cwd=vm_.get("cwd"), output_loglevel="info"
    )

    if machine == "":  # we were called using the default machine
        for line in ret["stdout"].split("\n"):  # find its actual Vagrant name
            if line.startswith("==>"):
                machine = line.split()[1].rstrip(":")
                vm_["machine"] = machine
                _update_vm_info(name, vm_)  # and remember the true name
                break

    if ret["retcode"] == 0:
        return f'Started "{name}" using Vagrant machine "{machine}".'
    return False


def shutdown(name):
    """
    Send a soft shutdown (vagrant halt) signal to the named vm.

    This does the same thing as vagrant.stop. Other-VM control
    modules use "stop" and "shutdown" to differentiate between
    hard and soft shutdowns.

    CLI Example:

    .. code-block:: bash

        salt <host> vagrant.shutdown <salt_id>
    """
    return stop(name)


def stop(name):
    """
    Hard shutdown the virtual machine. (vagrant halt)

    CLI Example:

    .. code-block:: bash

        salt <host> vagrant.stop <salt_id>
    """
    vm_ = get_vm_info(name)
    machine = vm_["machine"]

    cmd = f"vagrant halt {machine}"
    ret = __salt__["cmd.retcode"](cmd, runas=vm_.get("runas"), cwd=vm_.get("cwd"))
    return ret == 0


def pause(name):
    """
    Pause (vagrant suspend) the named VM.

    CLI Example:

    .. code-block:: bash

        salt <host> vagrant.pause <salt_id>
    """
    vm_ = get_vm_info(name)
    machine = vm_["machine"]

    cmd = f"vagrant suspend {machine}"
    ret = __salt__["cmd.retcode"](cmd, runas=vm_.get("runas"), cwd=vm_.get("cwd"))
    return ret == 0


def reboot(name, provision=False):
    """
    Reboot a VM. (vagrant reload)

    CLI Example:

    .. code-block:: bash

        salt <host> vagrant.reboot <salt_id> provision=True

    :param name: The salt_id name you will use to control this VM
    :param provision: (False) also re-run the Vagrant provisioning scripts.
    """
    vm_ = get_vm_info(name)
    machine = vm_["machine"]
    prov = "--provision" if provision else ""

    cmd = f"vagrant reload {machine} {prov}"
    ret = __salt__["cmd.retcode"](cmd, runas=vm_.get("runas"), cwd=vm_.get("cwd"))
    return ret == 0


def destroy(name):
    """
    Destroy and delete a virtual machine. (vagrant destroy -f)

    This also removes the salt_id name defined by vagrant.init.

    CLI Example:

    .. code-block:: bash

        salt <host> vagrant.destroy <salt_id>
    """
    vm_ = get_vm_info(name)
    machine = vm_["machine"]

    cmd = f"vagrant destroy -f {machine}"

    ret = __salt__["cmd.run_all"](
        cmd, runas=vm_.get("runas"), cwd=vm_.get("cwd"), output_loglevel="info"
    )
    if ret["retcode"] == 0:
        _erase_vm_info(name)
        return f"Destroyed VM {name}"
    return False


def get_ssh_config(name, network_mask="", get_private_key=False):
    r"""
    Retrieve hints of how you might connect to a Vagrant VM.

    :param name: the salt_id of the machine
    :param network_mask: a CIDR mask to search for the VM's address
    :param get_private_key: (default: False) return the key used for ssh login
    :return: a dict of ssh login information for the VM

    CLI Example:

    .. code-block:: bash

        salt <host> vagrant.get_ssh_config <salt_id>
        salt my_laptop vagrant.get_ssh_config quail1 network_mask=10.0.0.0/8 get_private_key=True

    The returned dictionary contains:

    - key_filename:  the name of the private key file on the VM host computer
    - ssh_username:  the username to be used to log in to the VM
    - ssh_host:  the IP address used to log in to the VM.  (This will usually be `127.0.0.1`)
    - ssh_port:  the TCP port used to log in to the VM.  (This will often be `2222`)
    - \[ip_address:\]  (if `network_mask` is defined. see below)
    - \[private_key:\]  (if `get_private_key` is True) the private key for ssh_username

    About `network_mask`:

    Vagrant usually uses a redirected TCP port on its host computer to log in to a VM using ssh.
    This redirected port and its IP address are "ssh_port" and "ssh_host".  The ssh_host is
    usually the localhost (127.0.0.1).
    This makes it impossible for a third machine (such as a salt-cloud master) to contact the VM
    unless the VM has another network interface defined.  You will usually want a bridged network
    defined by having a `config.vm.network "public_network"` statement in your `Vagrantfile`.

    The IP address of the bridged adapter will typically be assigned by DHCP and unknown to you,
    but you should be able to determine what IP network the address will be chosen from.
    If you enter a CIDR network mask, Salt will attempt to find the VM's address for you.
    The host machine will send an "ifconfig" command to the VM (using ssh to `ssh_host`:`ssh_port`)
    and return the IP address of the first interface it can find which matches your mask.
    """
    vm_ = get_vm_info(name)

    ssh_config = _vagrant_ssh_config(vm_)

    try:
        ans = {
            "key_filename": ssh_config["IdentityFile"],
            "ssh_username": ssh_config["User"],
            "ssh_host": ssh_config["HostName"],
            "ssh_port": ssh_config["Port"],
        }

    except KeyError:
        raise CommandExecutionError(
            "Insufficient SSH information to contact VM {}. Is it running?".format(
                vm_.get("machine", "(default)")
            )
        )

    if network_mask:
        #  ask the new VM to report its network address
        command = (
            "ssh -i {IdentityFile} -p {Port} "
            "-oStrictHostKeyChecking={StrictHostKeyChecking} "
            "-oUserKnownHostsFile={UserKnownHostsFile} "
            "-oControlPath=none "
            "{User}@{HostName} ifconfig".format(**ssh_config)
        )

        log.info("Trying ssh -p %(Port)s %(User)s@%(HostName)s ifconfig", ssh_config)
        reply = __salt__["cmd.shell"](command)
        log.info("--->\n%s", reply)
        target_network_range = ipaddress.ip_network(network_mask, strict=False)

        for line in reply.split("\n"):
            try:  # try to find a bridged network address
                # the lines we are looking for appear like:
                #    "inet addr:10.124.31.185  Bcast:10.124.31.255  Mask:255.255.248.0"
                # or "inet6 addr: fe80::a00:27ff:fe04:7aac/64 Scope:Link"
                tokens = line.replace(
                    "addr:", "", 1
                ).split()  # remove "addr:" if it exists, then split
                found_address = None
                if "inet" in tokens:
                    nxt = tokens.index("inet") + 1
                    found_address = ipaddress.ip_address(tokens[nxt])
                elif "inet6" in tokens:
                    nxt = tokens.index("inet6") + 1
                    found_address = ipaddress.ip_address(tokens[nxt].split("/")[0])
                if found_address in target_network_range:
                    ans["ip_address"] = str(found_address)
                    break  # we have located a good matching address
            except (IndexError, AttributeError, TypeError):
                pass  # all syntax and type errors loop here
                # falling out if the loop leaves us remembering the last candidate
        log.info(
            "Network IP address in %s detected as: %s",
            target_network_range,
            ans.get("ip_address", "(not found)"),
        )

    if get_private_key:
        # retrieve the Vagrant private key from the host
        try:
            with salt.utils.files.fopen(ssh_config["IdentityFile"]) as pks:
                ans["private_key"] = salt.utils.stringutils.to_unicode(pks.read())
        except OSError as e:
            raise CommandExecutionError(
                f"Error processing Vagrant private key file: {e}"
            )
    return ans



"""
Unit tests for the boto_cloudfront state module.
"""

import copy
import textwrap

import pytest

import salt.loader
import salt.states.boto_cloudfront as boto_cloudfront
from tests.support.mock import MagicMock, patch


@pytest.fixture
def configure_loader_modules(minion_opts):
    utils = salt.loader.utils(
        minion_opts,
        whitelist=[
            "boto3",
            "dictdiffer",
            "yaml",
            "args",
            "systemd",
            "path",
            "platform",
        ],
        context={},
    )
    return {boto_cloudfront: {"__utils__": utils}}


def base_ret_with(extra_ret):
    name = "my_distribution"
    base_ret = {"name": name, "changes": {}}
    new_ret = copy.deepcopy(base_ret)
    new_ret.update(extra_ret)
    return new_ret


def test_present_distribution_retrieval_error():
    """
    Test for boto_cloudfront.present when we cannot get the distribution.
    """
    name = "my_distribution"
    base_ret = {"name": name, "changes": {}}
    # Most attributes elided since there are so many required ones
    config = {"Enabled": True, "HttpVersion": "http2"}
    tags = {"test_tag1": "value1"}
    mock_get = MagicMock(return_value={"error": "get_distribution error"})
    with patch.multiple(
        boto_cloudfront,
        __salt__={"boto_cloudfront.get_distribution": mock_get},
        __opts__={"test": False},
    ):
        comment = "Error checking distribution {0}: get_distribution error"
        assert boto_cloudfront.present(name, config, tags) == base_ret_with(
            {"result": False, "comment": comment.format(name)}
        )


def test_present_from_scratch():
    name = "my_distribution"
    base_ret = {"name": name, "changes": {}}
    config = {"Enabled": True, "HttpVersion": "http2"}
    tags = {"test_tag1": "value1"}
    mock_get = MagicMock(return_value={"result": None})

    with patch.multiple(
        boto_cloudfront,
        __salt__={"boto_cloudfront.get_distribution": mock_get},
        __opts__={"test": True},
    ):
        comment = f"Distribution {name} set for creation."
        assert boto_cloudfront.present(name, config, tags) == base_ret_with(
            {"result": None, "comment": comment, "changes": {"old": None, "new": name}}
        )

    mock_create_failure = MagicMock(return_value={"error": "create error"})
    with patch.multiple(
        boto_cloudfront,
        __salt__={
            "boto_cloudfront.get_distribution": mock_get,
            "boto_cloudfront.create_distribution": mock_create_failure,
        },
        __opts__={"test": False},
    ):
        comment = "Error creating distribution {0}: create error"
        assert boto_cloudfront.present(name, config, tags) == base_ret_with(
            {"result": False, "comment": comment.format(name)}
        )

    mock_create_success = MagicMock(return_value={"result": True})
    with patch.multiple(
        boto_cloudfront,
        __salt__={
            "boto_cloudfront.get_distribution": mock_get,
            "boto_cloudfront.create_distribution": mock_create_success,
        },
        __opts__={"test": False},
    ):
        comment = "Created distribution {0}."
        assert boto_cloudfront.present(name, config, tags) == base_ret_with(
            {
                "result": True,
                "comment": comment.format(name),
                "changes": {"old": None, "new": name},
            }
        )


def test_present_correct_state():
    name = "my_distribution"
    base_ret = {"name": name, "changes": {}}
    # Most attributes elided since there are so many required ones
    config = {"Enabled": True, "HttpVersion": "http2"}
    tags = {"test_tag1": "value1"}

    mock_get = MagicMock(
        return_value={
            "result": {
                "distribution": {"DistributionConfig": config},
                "tags": tags,
                "etag": "test etag",
            }
        }
    )
    with patch.multiple(
        boto_cloudfront,
        __salt__={"boto_cloudfront.get_distribution": mock_get},
        __opts__={"test": False},
    ):
        comment = "Distribution {0} has correct config."
        assert boto_cloudfront.present(name, config, tags) == base_ret_with(
            {"result": True, "comment": comment.format(name)}
        )


def test_present_update_config_and_tags():
    name = "my_distribution"
    base_ret = {"name": name, "changes": {}}
    config = {"Enabled": True, "HttpVersion": "http2"}
    tags = {"test_tag1": "value1"}
    mock_get = MagicMock(
        return_value={
            "result": {
                "distribution": {
                    "DistributionConfig": {
                        "Enabled": False,
                        "Comment": "to be removed",
                    }
                },
                "tags": {"bad existing tag": "also to be removed"},
                "etag": "test etag",
            }
        }
    )

    diff = textwrap.dedent(
        """\
        ---
        +++
        @@ -1,5 +1,5 @@
         config:
        -  Comment: to be removed
        -  Enabled: false
        +  Enabled: true
        +  HttpVersion: http2
         tags:
        -  bad existing tag: also to be removed
        +  test_tag1: value1

    """
    ).splitlines()
    # Difflib adds a trailing space after the +++/--- lines,
    # programatically add them back here. Having them in the test file
    # itself is not feasible since a few popular plugins for vim will
    # remove trailing whitespace.
    for idx in (0, 1):
        diff[idx] += " "
    diff = "\n".join(diff)

    with patch.multiple(
        boto_cloudfront,
        __salt__={"boto_cloudfront.get_distribution": mock_get},
        __opts__={"test": True},
    ):
        header = f"Distribution {name} set for new config:"
        assert boto_cloudfront.present(name, config, tags) == base_ret_with(
            {
                "result": None,
                "comment": "\n".join([header, diff]),
                "changes": {"diff": diff},
            }
        )

    mock_update_failure = MagicMock(return_value={"error": "update error"})
    with patch.multiple(
        boto_cloudfront,
        __salt__={
            "boto_cloudfront.get_distribution": mock_get,
            "boto_cloudfront.update_distribution": mock_update_failure,
        },
        __opts__={"test": False},
    ):
        comment = "Error updating distribution {0}: update error"
        assert boto_cloudfront.present(name, config, tags) == base_ret_with(
            {"result": False, "comment": comment.format(name)}
        )

    mock_update_success = MagicMock(return_value={"result": True})
    with patch.multiple(
        boto_cloudfront,
        __salt__={
            "boto_cloudfront.get_distribution": mock_get,
            "boto_cloudfront.update_distribution": mock_update_success,
        },
        __opts__={"test": False},
    ):
        assert boto_cloudfront.present(name, config, tags) == base_ret_with(
            {
                "result": True,
                "comment": f"Updated distribution {name}.",
                "changes": {"diff": diff},
            }
        )



"""
    :codeauthor: Anthony Shaw <anthonyshaw@apache.org>

    TestCase for salt.modules.servicenow
"""

import pytest

import salt.modules.servicenow as servicenow
from tests.support.mock import MagicMock


class MockServiceNowClient:
    def __init__(self, instance_name, username, password):
        pass

    def get(self, query):
        return [{"query_size": len(query), "query_value": query}]


@pytest.fixture
def configure_loader_modules():
    module_globals = {
        "Client": MockServiceNowClient,
        "__salt__": {
            "config.option": MagicMock(
                return_value={
                    "instance_name": "test",
                    "username": "mr_test",
                    "password": "test123",
                }
            )
        },
    }
    if servicenow.HAS_LIBS is False:
        module_globals["sys.modules"] = {"servicenow_rest": MagicMock()}
        module_globals["sys.modules"][
            "servicenow_rest"
        ].api.Client = MockServiceNowClient
    return {servicenow: module_globals}


def test_module_creation():
    client = servicenow._get_client()
    assert client is not None


def test_non_structured_query():
    result = servicenow.non_structured_query("tests", "role=web")
    assert result is not None
    assert result[0]["query_size"] == 8
    assert result[0]["query_value"] == "role=web"


def test_non_structured_query_kwarg():
    result = servicenow.non_structured_query("tests", role="web")
    assert result is not None
    assert result[0]["query_size"] == 8
    assert result[0]["query_value"] == "role=web"


def test_non_structured_query_kwarg_multi():
    result = servicenow.non_structured_query("tests", role="web", type="computer")
    assert result is not None
    assert result[0]["query_size"] == 22



"""
    tests.unit.test_pytest_pass_fail
    ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

    Some tests to make sure our pytest usage doesn't break regular pytest behviour
"""

import pytest


@pytest.mark.xfail(
    strict=True,
    reason="This test should always fail. If it passes, we messed up pytest",
)
def test_should_always_fail():
    assert False


def test_should_always_pass():
    assert True
